/**
 * @(#)Communicator.java
 *
 * Package: simu.net
 *
 * This file contains the Communicator class, used by any
 * thread that needs to communicate with the Simulator
 * Control Program (SCP).
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970320   Andre Campeau   Creation.
 *
 */


package simu.net;


import java.net.Socket;
import java.net.ServerSocket;
import java.net.InetAddress;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.util.Hashtable;


/**
 * @version        1.0 20-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The Communicator class is used by objects running in the
 * simulated network environment to communicate with the
 * Simulator Control Program (SCP).  In most cases, the objects
 * would be Simulated Network Elements (SNEs) which simulate
 * the behaviour of real network elements, such as printers and
 * routers.  The SNEs are Java programs executing on computers
 * which are part of the simulated network.
 * <p>
 * The communication with the Simulator Control Program is
 * accomplished using TCP socket connections.  This way, the
 * SCP can be accessed regardless of what computer it's
 * running on.  This allows the simulated network environment
 * to be extended to many computers, with SNEs running on each.
 * <p>
 * The Communicator class may be use in one of three ways:
 * <OL>
 * <LI> Spawned to interact with a dispatcher and obtain the
 *      anonymous port number to use when subsequently connecting
 *      to the SCPCommunicator.
 * <LI> Spawned by an object (such as a SNE) to be permanently
 *      listening (as a daemon) for communication.
 * <LI> Spawned by an object (such as a SNE) to communicate with
 *      the SCP, after which the Communicator thread terminates.
 * </OL>
 * <p>
 * Once the Communicator is constructed, the guts of it (the parts
 * that send and receive data) are spawned as a separate thread (a
 * daemon thread in case 1 above).  Receive and Send methods exist
 * to handle the sending and receiving of data once the socket
 * connections have been established.
 * <p>
 * The Communicator communicates with other objects on the
 * network using an established protocol.  Any incoming messages
 * that do not adhere to the protocol are not processed.  A
 * similar protocol is used by the SCPCommunicator class.
 * <p>
 * Valid protocol messages that the Communicator can receive:<BR>
 * <UL>
 * <LI> "SHIP {key} {value}"
 * <LI> "SHIP REMOVE {key}"
 * <LI> "RECEIVE {key}"
 * <LI> "RECEIVE REMOTE ALL"
 * </UL>
 * <p>
 * Valid protocol messages that the Communicator can send:<BR>
 * <UL>
 * <LI> "RECEIVE LOCAL {SNEid} {key}"
 * <LI> "RECEIVE LINKS {fromSNEid} {toSNEid}"
 * <LI> "SHIP LINKS {fromSNEid} {toSNEid} {linkStatus}"
 * </UL>
 * <p>
 * @see simu.net.SCPCommunicator
 */
public class Communicator extends Thread
{
  // The following class variables are set by the various
  // constructors, so that the spawned thread has access
  // to required data.  For more information, see the
  // constructor method descriptions below.
  private int listenerPort = 0;
  private ServerSocket listener = null;
  private String sendData = null;
  private String receiveData[] = null;
  private Hashtable localTable = null;
  private int sendPort = 0;
  private int sendPortArray[] = new int[1];
  private String sendIP = null;

  // ----------------------------------------------------------------------------

  /**
   * This Communicator constructor interacts with a dispatcher.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The dispatcher will accept a socket connection,
   *                 pass the next free port number through the
   *                 socket, then close the socket connection.
   *                 The purpose of this thread is to obtain the
   *                 next free port number from a dispatcher and
   *                 return it to the caller, after which it
   *                 terminates.
   *
   * @param SendIP    The IP address of the remote object.
   * @param SendPort  The listener port of the remote object.
   */
  public Communicator (String SendIP, int[] SendPort)
  {
    super ("Communicator");

    // Initialize the class data.
    sendIP = SendIP;
    sendPortArray = SendPort;

    // Set the thread as a daemon and start it.
    this.start();
  }

  // ----------------------------------------------------------------------------

  /**
   * This Communicator constructor spawns a listener daemon thread.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The Communicator object is constructed, and
   *                 a listener thread is spawned.  The thread acts
   *                 as a daemon and listens continuously for
   *                 incoming data (and possibly sends data as a
   *                 reply).  The thread is terminated when the
   *                 all non-daemon threads in the JVM terminate.
   *
   * @param LocalTable    A pointer to the local Hashtable.  Note that
   *                      if a SNE is constructing the Communicator,
   *                      the localTable may be the MIB.  In general,
   *                      however, there is no guarantee that this
   *                      will be the case.
   * @param ListenerPort  The port the daemon listens on.
   */
  public Communicator (Hashtable LocalTable, int ListenerPort)
  {
    super ("Communicator");

    // Initialize the class data.
    listenerPort = ListenerPort;
    localTable = LocalTable;

    // Set the thread as a daemon and start it.
    this.setDaemon (true);
    this.start();
  }

  // ----------------------------------------------------------------------------

  /**
   * This Communicator constructor spawns a temporary sender thread.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The Communicator object is constructed, and
   *                 a temporary sending thread is spawned.  Data
   *                 is sent to the specified IP address and port,
   *                 after which data may be received over the
   *                 same connection.  Any received data is stored
   *                 in a String array of size 1.
   *
   * @param ReceiveData[]  The String (of size 1) where received
   *                       data is stored.
   * @param SendData       The data to be sent to the remote object.
   * @param SendIP         The IP address of the remote object.
   * @param SendPort       The listener port of the remote object.
   */
  public Communicator (String[] ReceiveData, String SendData, String SendIP, int SendPort)
  {
    super ("Communicator");

    // Initialize the class data.
    receiveData = ReceiveData;
    sendData = SendData;
    sendIP = SendIP;
    sendPort = SendPort;

    // Start the thread.
    this.start();
  }

  // ----------------------------------------------------------------------------

  /**
   * The Receive method reads data from the connected socket, and
   * sends a reply as appropriate.
   * <p>
   * Preconditions:  The listener thread has accepted a connection
   *                 request from a remote object, and the input
   *                 and output streams have been attached to the
   *                 connection.<p>
   * Postconditions: Data is received from the remote object over
   *                 the socket connection.  If the remote object
   *                 expects a reply, the localTable is queried to
   *                 get the data, which is sent over the connection.
   *
   * @param dis  The data input stream attached to the socket
   *             connection.
   * @param ps   The output printstream attached to the socket
   *             connection.
   */
  public void Receive (DataInputStream dis, PrintStream ps)
  {
    // Read data from the socket.
    try
    {
      String receiveData = dis.readLine();

      if (receiveData.startsWith ("RECEIVE"))
      // If the received data starts with "RECEIVE", then we will
      // have to process the rest of the received data and send
      // some kind of reply.  See the valid protocol messages part
      // of the Communicator class definition for more info.
      {
        StringTokenizer st = new StringTokenizer (receiveData);
        st.nextToken();  // Throw away the mode token.

        if (receiveData.equals ("RECEIVE REMOTE ALL"))
        // If the other side wants all the data in the localTable,
        // send it over the link piece by piece.
        {
          Enumeration localTablekeys = localTable.keys();
          while (localTablekeys.hasMoreElements())
          {
            String key = (String)localTablekeys.nextElement();
            ps.println (key);
            String value = (String)localTable.get (key);
            ps.println (value);
          }
          ps.println ("END_OF_DATA");  // Tell other side there's no more data.
        }
        else if (st.countTokens() == 1)
        // If there are no more elements in the String, then the
        // other side wants us to send the value corresponding to
        // the key.
        {
          String key = st.nextToken();  // Get key from receiveData String.
          ps.println (localTable.get (key));  // Send the value directly from
                                              // the localTable.
        }
        else
        // This is an error condition.
        {
        }
      }
      else if (receiveData.startsWith ("SHIP"))
      // If the other side wants to send some data to us, get
      // the data (which must be links data, since that is
      // all that is supported, for now) and update the SNED.
      {
        StringTokenizer st = new StringTokenizer (receiveData);
        st.nextToken();  // Throw away the mode token.

        if (receiveData.startsWith ("SHIP REMOVE"))
        // If the other side wants to delete an entry in the localTable,
        // get the key and delete the entry.
        {
          st.nextToken();  // Throw away "REMOVE".
          String key = st.nextToken();
          localTable.remove (key);
        }
        else if (st.countTokens() >= 2)
        // If there are no more elements in the String, then the
        // other side wants us to add the key and value that it
        // sent to us to the localTable.
        {
          String key = st.nextToken();
          String value = st.nextToken();
          while (st.hasMoreTokens())
          // If the attribute has spaces in it, add them to the
          // value string.
          {
            value = value + " " + st.nextToken();
          }
          localTable.put (key, value);
        }
        else
        // This is an error condition.
        {
        }
      }
      else
      // This is an error condition.
      {
      }
    }
    catch (IOException ioe)
    {
    }
  }

  // ----------------------------------------------------------------------------

  /**
   * The Send method sends data out of the connected socket, and
   * waits for a reply as appropriate.
   * <p>
   * Preconditions:  The thread has made a connection to the remote
   *                 object, and the input and output streams have
   *                 been attached to the connection.<p>
   * Postconditions: Data is sent to the remote object over the
   *                 socket connection.  If the remote object
   *                 expects a reply, the localTable is queried to
   *                 get the data, which is sent over the connection.
   *
   * @param dis  The data input stream attached to the socket
   *             connection.
   * @param ps   The output printstream attached to the socket
   *             connection.
   */
  public void Send (String[] receiveData, String sendData,
                    DataInputStream dis, PrintStream ps)
  {
    try
    {
      if (sendData.startsWith ("RECEIVE LOCAL"))
      {
        // Send the command to retrieve the data.
        ps.println (sendData);

        // Now wait for the response to appear.
        receiveData[0] = dis.readLine();
      }
      else if (sendData.startsWith ("RECEIVE LINKS"))
      {
        // Send the command to retrieve the data.
        ps.println (sendData);

        // Now wait for the response to appear.
        receiveData[0] = dis.readLine();
      }
      else if (sendData.startsWith ("SHIP LINKS"))
      {
        // Send the data across the connection.
        ps.println (sendData);
      }
      else
      // This is an error condition.
      {
      }
    }
    catch (IOException ioe)
    {
    }

  }

  // ----------------------------------------------------------------------------

  /**
   * The run method contains the code that will be run by the
   * Communicator thread.
   */
  public void run()
  {
    try
    {
      if (sendData == null && localTable == null)
      // If we have no data to send and we are not expecting to
      // receive any data (because localTable is null), we must be
      // trying to establish a connection with the Dispatcher.
      {
        // Open a socket connection.
        Socket connection = new Socket (InetAddress.getByName (sendIP), sendPortArray[0]);
        System.out.println ("Sender: Connection made to Dispatcher.");

        // set up the streams
        InputStream inStream = connection.getInputStream();
        DataInputStream dis = new DataInputStream (inStream);

        // Set the sendPortArray to the value of the anonymous port.
        sendPortArray[0] = Integer.parseInt (dis.readLine());

        // close the input stream and server socket
        dis.close();
        connection.close();
        System.out.println ("Sender: Dispatcher socket closed.");
      }
      else if (sendData == null)
      // If there is no data to send, the purpose of the thread
      // must be to receive data.
      {
        while (true)
        {
          listener = new ServerSocket (listenerPort);
          System.out.println ("Communicator: Server Socket waiting on port " +
                               listenerPort + " ...");

          // Wait for a connection.
          Socket connection = listener.accept();
          System.out.println ("Listener: Connection accepted.");

          // Set up the streams.
          InputStream inStream = connection.getInputStream();
          OutputStream outStream = connection.getOutputStream();
          DataInputStream dis = new DataInputStream (inStream);
          PrintStream ps = new PrintStream (outStream);

          // Read data from the socket, do any required processing,
          // then close the connection.
          this.Receive (dis, ps);

          // close the input stream and server socket
          dis.close();
          ps.close();
          listener.close();
          System.out.println ("Listener: Server socket closed.");
        }
      }
      else
      // If we want to ship data, simply take the data and send
      // it to the remote location.  If the data we are shipping
      // is preceded by a "RECEIVE" keyword, we will have to
      // wait to receive a reply from the other side.  The thread
      // will terminate once the data is sent and any replies are
      // processed.
      {
        // Open a socket connection.
        Socket connection = new Socket (InetAddress.getByName (sendIP), sendPort);
        System.out.println ("Sender: Connection made.");

        // set up the streams
        InputStream inStream = connection.getInputStream();
        OutputStream outStream = connection.getOutputStream();
        DataInputStream dis = new DataInputStream (inStream);
        PrintStream ps = new PrintStream (outStream);

        // Send the data and do any required processing.
        this.Send (receiveData, sendData, dis, ps);

        // close the input stream and server socket
        dis.close();
        ps.close();
        connection.close();
        System.out.println ("Sender: Server socket closed.");
      }
    }
    catch (IOException ioe)
    {
      ioe.printStackTrace();
    }
    System.out.println ("Communicator thread termination");
  }

} // End class Communicator
