/**
 * @(#)Dispatcher.java
 *
 * Package: simu.net
 *
 * This file contains the Dispatcher class, used by the
 * Simulator Control Program (SCP) to communicate with
 * other programs on the network.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970316   Andre Campeau   Creation.
 *
 */


package simu.net;


import java.awt.TextArea;
import java.net.ServerSocket;
import java.net.Socket;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.IOException;
import simu.sned.sned;
import simu.net.SCPCommunicator;


/**
 * @version        1.0 16-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The Dispatcher class is spawned as a dameon thread by
 * the Simulator Control Program (SCP) which listens for
 * communication connection requests from remote objects on
 * the network.  When a TCP socket connection is made,
 * the Dispatcher accepts it, creates a new ServerSocket
 * on an anonymous port, gets the port number from the
 * new ServerSocket and sends this port number to the remote
 * object, after which it terminates the connection and
 * proceeds to listen again for communication connection
 * requests.
 * <p>
 * The new ServerSocket is passed to a SCPCommunicator,
 * which listens for a subsequent connection request on
 * the new port.  This allows the dispatcher to listen
 * for future communication requests while another thread
 * (the SCPCommunicator in this case) processes pending
 * requests.
 * <p>
 * @see simu.net.SCPCommunicator
 */
public class Dispatcher extends Thread
{
  // The following class variables are set by the
  // constructor, so that the spawned thread has access
  // to required data.  For more information, see the
  // constructor method description below.
  private sned theSned;
  private int listenerPort;
  private java.awt.TextArea textArea_StatusMessages;

  // ----------------------------------------------------------------------------

  /**
   * This Dispatcher constructor spawns a listener daemon thread.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The listener daemon thread is spawned.
   *
   * @param TheSned       A pointer to the Simulated Network
   *                      Element Database (SNED).
   * @param ListenerPort  The well-known port that the Dispatcher
   *                      thread will listen on.
   * @param TextArea_StatusMessages
   *                      The Dispatcher will occasionally print
   *                      status messages to the SCP status
   *                      messages text area, so a pointer to this
   *                      is required.
   */
  public Dispatcher (sned TheSned, int ListenerPort,
                     java.awt.TextArea TextArea_StatusMessages)
  {
    listenerPort = ListenerPort;
    theSned = TheSned;
    textArea_StatusMessages = TextArea_StatusMessages;

    this.setDaemon (true);
    this.start();
  }

  // ----------------------------------------------------------------------------

  /**
   * The displayStatusMessage method prints status messages to the
   * Simulator Control Program (SCP) status messages text area.
   * Note that this method is private, so it will not appear in
   * the Javadoc generated HTML documentation pages.
   * <p>
   * Preconditions:  TextArea_StatusMessages points to the SCP
   *                 status messages text area.<p>
   * Postconditions: The message is printed to the SCP status
   *                 messages text area, followed by a linefeed
   *                 character.
   *
   * @param Message  The message to print in the SCP status
   *                 messages text area.
   */
  private synchronized void displayStatusMessage (String Message)
  {
    textArea_StatusMessages.appendText (Message + "\n");
  }

  // ----------------------------------------------------------------------------

  /**
   * The run method contains the code that will be run by the
   * Dispatcher thread.
   */
  public void run()
  {
    try
    {
      while (true)
      {
        // Create the Dispatcher ServerSocket and wait for a
        // TCP connection request.
        ServerSocket dispatcher = new ServerSocket (listenerPort);
        Socket connection = dispatcher.accept();

        // Once the connection has been made, set up the streams.
        OutputStream outStream = connection.getOutputStream();
        PrintStream ps = new PrintStream (outStream);

        // Create the new ServerSocket with anonymous port to
        // farm out to the SCPCommunicator.
        ServerSocket listener = new ServerSocket (0);

        // Get the port number of the anonymous ServerSocket and
        // send it out of the Dispatcher connection socket.
        ps.println (listener.getLocalPort());

        // Close the output stream and server socket.
        ps.close();
        dispatcher.close();

        displayStatusMessage ("Dispatcher: Spawning SCPCommunicator on port " +
                             listener.getLocalPort());

        // Spawn the SCPCommunicator thread and pass it the SNED and
        // a pointer to the listener ServerSocket created earlier.
        SCPCommunicator theComm = new SCPCommunicator (theSned, listener);
      }
    }
    catch (IOException ioe) {}
  }
}