/**
 * @(#)SCPCommunicator.java
 *
 * Package: simu.net
 *
 * This file contains the SCPCommunicator class, used by
 * the Simulator Control Program (SCP) to communicate with
 * other programs on the network.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970316   Andre Campeau   Creation.
 *
 */


package simu.net;


import java.net.Socket;
import java.net.ServerSocket;
import java.net.InetAddress;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.Hashtable;
import java.util.StringTokenizer;
import simu.sned.sned;
import simu.util.Semaphore;


/**
 * @version        1.0 16-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The SCPCommunicator class is used by the Simulator Control
 * program (SCP) to communicate with remote objects on the simulated
 * network environment.  The simulated network environment is
 * composed of Simulated Network Elements (SNEs) which simulate
 * the behaviour of real network elements, such as printers and
 * routers.  The SNEs are Java programs executing on the local
 * or remote computers.
 * <p>
 * The communication with the remote objects (usually SNEs) is
 * accomplished using TCP socket connections.  This way, the
 * remote objects can be accessed regardless of whether they
 * are on the local computer (localhost) or a remote computer
 * on the network.  This allows the simulated network environment
 * to be extended to many computers, with SNEs running on each.
 * <p>
 * The SCPCommunicator class may be use in one of three ways:
 * <OL>
 * <LI> Spawned by the SCP to be permanently listening (as a
 *      daemon) for communication.
 * <LI> Spawned by the Dispatcher to listen for one and only one
 *      incoming message, after which the SCPCommunicator
 *      thread terminates.
 * <LI> Spawned by the SCP to communicate with a remote object,
 *      such as a SNE, after which the SCPCommunicator thread
 *      terminates.
 * </OL>
 * <p>
 * Once the SCPCommunicator is constructed, the guts of it (the parts
 * that send and receive data) are spawned as a separate thread (a
 * daemon thread in case 1 above).  Receive and Send methods exist
 * to handle the sending and receiving of data once the socket
 * connections have been established.
 * <p>
 * The SCPCommunicator communicates with other objects on the
 * network using an established protocol.  Any incoming messages
 * that do not adhere to the protocol are not processed.  A
 * similar protocol is used by the Communicator class.
 * <p>
 * Valid protocol messages that the SCPCommunicator can receive:<BR>
 * <UL>
 * <LI> "RECEIVE LOCAL {SNEid} {key}"<BR>
 * <LI> "RECEIVE LINKS {fromSNEid} {toSNEid}"<BR>
 * <LI> "SHIP LINKS {fromSNEid} {toSNEid} {linkStatus}"<BR>
 * </UL>
 * <p>
 * Valid protocol messages that the SCPCommunicator can send:<BR>
 * <UL>
 * <LI> "SHIP {key} {value}"<BR>
 * <LI> "SHIP REMOVE {key}"<BR>
 * <LI> "RECEIVE {key}"<BR>
 * <LI> "RECEIVE REMOTE ALL"<BR>
 * </UL>
 * <p>
 * @see simu.net.Dispatcher
 * @see simu.net.Communicator
 */
public class SCPCommunicator extends Thread
{
  // The following class variables are set by the various
  // constructors, so that the spawned thread has access
  // to required data.  For more information, see the
  // constructor method descriptions below.
  private int listenerPort = 0;
  private ServerSocket listener = null;
  private String sendData = null;
  private Hashtable remoteData;
  private int sendPort = 0;
  private String sendIP = null;
  private sned theSned = null;
  private boolean anError = false;
  private Semaphore sem = new Semaphore();

  // ----------------------------------------------------------------------------

  /**
   * This SCPCommunicator constructor spawns a listener daemon thread.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The SCPCommunicator object is constructed, and
   *                 a listener thread is spawned.  The thread acts
   *                 as a daemon and listens continuously for
   *                 incoming data (and possibly sends data as a
   *                 reply).  The thread is terminated when the
   *                 all non-daemon threads in the JVM terminate.
   *
   * @param TheSned       A pointer to the SNE database, which is
   *                      created by the SCP.
   * @param ListenerPort  The port the daemon listens on.
   */
  public SCPCommunicator (sned TheSned, int ListenerPort)
  {
    super ("SCPCommunicator");

    // Initialize the class data.
    listenerPort = ListenerPort;
    theSned = TheSned;

    // Set the thread as a daemon and start it.
    this.setDaemon (true);
    this.start();
  }

  // ----------------------------------------------------------------------------

  /**
   * This SCPCommunicator constructor spawns a temporary listener thread.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The SCPCommunicator object is constructed, and
   *                 a listener thread is spawned.  The thread
   *                 listens until the short-term objective is
   *                 accomplished (i.e. receive remote data), then
   *                 it terminates.
   *
   * @param TheSned   A pointer to the SNE database, which is
   *                  created by the SCP.
   * @param Listener  A pointer to the ServerSocket to listen on.
   */
  public SCPCommunicator (sned TheSned, ServerSocket Listener)
  {
    super ("SCPCommunicator");

    // Initialize the class data.
    listener = Listener;
    theSned = TheSned;

    // Start the thread.
    this.start();
    sem.Wait();
  }

  // ----------------------------------------------------------------------------

  /**
   * This SCPCommunicator constructor spawns a temporary sender thread.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The SCPCommunicator object is constructed, and
   *                 a sending thread is spawned.  Data is sent to
   *                 the specified IP address and port, after which
   *                 data may be received over the same connection.
   *                 Any received data is stored in a Hashtable.
   *
   * @param RemoteData  The Hashtable where received remote data is
   *                    stored.
   * @param SendData    The data to be sent to the remote object.
   * @param SendIP      The IP address of the remote object.
   * @param SendPort    The listener port of the remote object.
   */
  public SCPCommunicator (Hashtable RemoteData, String SendData, String SendIP, int SendPort)
  {
    super ("SCPCommunicator");

    // Initialize the class data.
    remoteData = RemoteData;
    sendData = SendData;
    sendIP = SendIP;
    sendPort = SendPort;

    // Start the thread.
    this.start();
  }

 // ----------------------------------------------------------------------------

  /**
   * The isError method returns the error state of the
   * SCPCommunicator.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: If an error has been previsouly detected,
   *                 true is returned, otherwise false is returned.
   *                 An error is detected if data cannot be sent
   *                 to or received from the remote object that
   *                 the SCPCommunicator is connected to.
   */
  public boolean isError()
  {
    return anError;  // If an error has been detected, return true.
  }

  // ----------------------------------------------------------------------------

  /**
   * The Receive method reads data from the connected socket, and
   * sends a reply as appropriate.
   * <p>
   * Preconditions:  The listener thread has accepted a connection
   *                 request from a remote object, and the input
   *                 and output streams have been attached to the
   *                 connection.<p>
   * Postconditions: Data is received from the remote object over
   *                 the socket connection.  If the remote object
   *                 expects a reply, the SNED is queried to get
   *                 the data, which is sent over the connection.
   *
   * @param dis  The data input stream attached to the socket
   *             connection.
   * @param ps   The output printstream attached to the socket
   *             connection.
   */
  public void Receive (DataInputStream dis, PrintStream ps)
  {
    // Read data from the socket.
    try
    {
      String receiveData = dis.readLine();

      if (receiveData.startsWith ("RECEIVE"))
      // If the received data starts with "RECEIVE", then we will
      // have to process the rest of the received data and send
      // some kind of reply.  See the valid protocol messages part
      // of the SCPCommunicator class definition for more info.
      {
        StringTokenizer st = new StringTokenizer (receiveData);
        st.nextToken();  // Throw away the mode token.
        String type = st.nextToken();

        if (type.equals ("LOCAL"))
        // If the other side wants local data (i.e. not link data),
        // query the SNED and send the data.
        {
          String SNEid = st.nextToken();
          String key = st.nextToken();
          ps.println (key + " " + theSned.GetSNEData (SNEid, key));
        }
        else if (type.equals ("LINKS"))
        // If the other side wants link data, query the SNED and
        // send link status information.
        {
          String fromSNEid = st.nextToken();
          String toSNEid = st.nextToken();
          String linkStatus = theSned.GetLinkStatus (fromSNEid, toSNEid);

          if (linkStatus.length() == 0)
          // If there is no link, or one of the SNE's specified
          // doesn't exist, we want to tell the other side that
          // there is no such link.
          {
            linkStatus = "NO_SUCH_LINK";
          }
          ps.println (theSned.GetLinkStatus (fromSNEid, toSNEid));
        }
        else
        // This is an error condition.
        {
        }
      }
      else if (receiveData.startsWith ("SHIP"))
      // If the other side wants to send some data to us, get
      // the data (which must be links data, since that is
      // all that is supported, for now) and update the SNED.
      {
        StringTokenizer st = new StringTokenizer (receiveData);
        st.nextToken();  // Throw away the mode token.
        String type = st.nextToken();

        if (type.equals ("LINKS"))
        {
          String fromSNEid = st.nextToken();
          String toSNEid = st.nextToken();
          String linkStatus = st.nextToken();
          theSned.SetLinkStatus (fromSNEid, toSNEid, linkStatus);
        }
        else
        // This is an error condition.
        {
        }
      }
      else
      // This is an error condition.
      {
      }
    }
    catch (IOException ioe)
    {
    }
  }

  // ----------------------------------------------------------------------------

  /**
   * The Send method sends data out of the connected socket, and
   * waits for a reply as appropriate.
   * <p>
   * Preconditions:  The thread has made a connection to the remote
   *                 object, and the input and output streams have
   *                 been attached to the connection.<p>
   * Postconditions: Data is sent to the remote object over the
   *                 socket connection.  If the remote object
   *                 expects a reply, the SNED is queried to get
   *                 the data, which is sent over the connection.
   *
   * @param dis  The data input stream attached to the socket
   *             connection.
   * @param ps   The output printstream attached to the socket
   *             connection.
   */
  public void Send (Hashtable remoteData, String sendData,
                    DataInputStream dis, PrintStream ps)
  {
    try
    {
      if (sendData.equals ("RECEIVE REMOTE ALL"))
      {
        // Send the command to retrieve the data.
        ps.println (sendData);

        // Now wait for the response to appear.
        String receiveData = dis.readLine();
        while (! receiveData.equals ("END_OF_DATA"))
        {
          String key = receiveData;      // Set the key.
          receiveData = dis.readLine();  // Get the next data element.
          String value = receiveData;    // Set the value.
          remoteData.put (key, value);   // Put key + value in Hashtable.
          receiveData = dis.readLine();  // Get more data.
        }
      }
      else if (sendData.startsWith ("RECEIVE"))
      {
        StringTokenizer st = new StringTokenizer (sendData);
        st.nextToken();  // Throw away the mode token.
        String key = st.nextToken();

        // Send the command to retrieve the data.
        ps.println (sendData);

        // Now wait for the response to appear.
        String value = dis.readLine();

        remoteData.put (key, value);  // Put key + value in Hashtable.
      }
      else if (sendData.startsWith ("SHIP"))
      {
        // Send the data across the connection.
        ps.println (sendData);
      }
      else
      // This is an error condition.
      {
      }
    }
    catch (IOException ioe)
    {
    }
  }

  // ----------------------------------------------------------------------------

  /**
   * The run method contains the code that will be run by the
   * SCPCommunicator thread.
   */
  public void run()
  {
    try
    {
      if (sendData == null)
      // If there is no data to send, the purpose of the thread
      // must be to receive data.
      {
        boolean loopFlag = true;
        if (listenerPort == 0 && listener != null)
        // If the listener port is zero but the listener
        // ServerSocket was specified, then this thread is
        // not a daemon, so set loopFlag to false.
        {
          loopFlag = false;
        }
        if (listenerPort != 0 && listener == null)
        // If the listener port was specified, then this
        // thread is a daemon, so don't touch the loopFlag
        // and set up the listener ServerSocket.
        {
          listener = new ServerSocket (listenerPort);
        }
        do
        {
          System.out.println ("SCPCommunicator: Server Socket waiting on port " +
                               listener.getLocalPort() + " ...");

          // Wait for a connection.
          sem.Signal();
          Socket connection = listener.accept();
          System.out.println ("Listener: Connection accepted.");

          // Set up the streams.
          InputStream inStream = connection.getInputStream();
          OutputStream outStream = connection.getOutputStream();
          DataInputStream dis = new DataInputStream (inStream);
          PrintStream ps = new PrintStream (outStream);

          // Read data from the socket, do any required processing,
          // then close the connection.
          this.Receive (dis, ps);

          // close the input stream and server socket
          dis.close();
          ps.close();
          listener.close();
          System.out.println ("Listener: Server socket closed.");

          if (loopFlag)
          // If we are looping, we need to create a new ServerSocket
          // on the listenerPort.
          {
            listener = new ServerSocket (listenerPort);
          }
        } while (loopFlag);
      }

      else
      // If we want to ship data, simply take the data and send
      // it to the remote location.  If the data we are shipping
      // is preceded by a "RECEIVE" keyword, we will have to
      // wait to receive a reply from the other side.  The thread
      // will terminate once the data is sent and any replies are
      // processed.
      {
        remoteData.clear();  // Clear the receive data Hashtable.

        // Open a socket connection.
        Socket connection = new Socket (InetAddress.getByName (sendIP), sendPort);
        System.out.println ("Sender: Connection made.");

        // set up the streams
        InputStream inStream = connection.getInputStream();
        OutputStream outStream = connection.getOutputStream();
        DataInputStream dis = new DataInputStream (inStream);
        PrintStream ps = new PrintStream (outStream);

        // Send the data and do any required processing.
        this.Send (remoteData, sendData, dis, ps);

        // close the input stream and server socket
        dis.close();
        ps.close();
        connection.close();
        System.out.println ("Sender: Server socket closed.");
      }
    }
    catch (IOException ioe)
    {
      anError = true;
    }
    System.out.println ("SCPCommunicator thread termination");
  }

} // End class SCPCommunicator
