/**
 * @(#)SCP.java
 *
 * Simulator Control Program (SCP)
 *
 * This file contains the SCP class, which is the Simulator
 * Control Program applet used to control the simulated network
 * environment.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970301   Andre Campeau   Creation.
 *
 */


package simu.scp;


import java.awt.*;
import java.applet.*;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;
import simu.sned.sned;
import simu.net.Dispatcher;
import simu.net.SCPCommunicator;


/**
 * @version        1.0 1-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The SCP (Simulator Control Program) class is used to control the
 * simulated network environment.  The simulated network environment
 * is composed of a number of simulated network elements (SNEs),
 * which simulate the behaviour of real hardware devices such as
 * printers, routers or networked workstations, and house the mobile
 * code infrastructure.  Mobile code will migrate to and perform
 * tasks in SNEs just as they would on a real network.
 * <p>
 * The SCP is a basic extension of the java.applet.Applet class.
 * <p>
 * The SCP is responsible for creating, linking, destroying,
 * shipping code to and modifying data within these SNEs.
 * Multiple SNEs can be created on a single workstation, or on
 * many workstations.  Currently a user must run the SNE manually
 * on a workstation and then manually register it with the SCP.
 * <p>
 * In the future, an anchor daemon could be responsible for creating
 * the SNE at a request from the SCP, after which the SCP would
 * manage the SNE directly.  The anchor daemon would be a Java
 * program running on another computer which would respond to
 * create and destroy requests from the SCP.
 * <p>
 * SNEs can be linked together, thus simulating real network connections.
 * The SCP is responsible for managing the links between SNEs.
 * <p>
 * The SCP contains an interface component, which accepts input from
 * a user and displays simulation data. It also contains data elements,
 * such as global simulator data.
 * <p>
 * The user interface functions that are required for the SCP are:<UL>
 * <LI> create SNEs
 * <LI> manage SNE links
 * <LI> destroy SNEs
 * <LI> modify SNE data</UL>
 * <p>
 * The SCP uses something called a SNED, which stands for Simulated
 * Network Element Database, to store global non-remote data, which
 * consists of SNE registration information (IP, port, etc.) and simulated
 * link information.  For more information about the SNED, refer to the
 * file sned.java, which contains the sned class used to implement the SNED.
 */
public class SCP extends Applet {

  // Define constants.
  int TAB_KEY = 9;  // Define a constant corresponding to the Tab key ASCII value.

  // Specify the well-known port that the SCP will listen on.
  int listenerPort = 2200;

  // Define global data.
  sned theSned;  // Define the Simulated Network Element Database.

  // Define the Bell audio clip object.
  AudioClip Bell;

  // Define the Hashtable where the remote attributes will be stored
  // after retrieving them from the SNE.
  Hashtable RemoteAttributes;

/*  ------------------------------------------------------------------------------
    The following methods are custom-made to support the SCP.
 */

  /**
   * This displayStatusMessage method displays a message in the
   * Status Messages text area, followed by a linefeed.
   *
   * @param message  The message to be displayed in the Status
   *                 Messages text area.
   */
  public synchronized void displayStatusMessage (String Message)
  {
    textArea_StatusMessages.appendText (Message + "\n");
  }

  /**
   * This displayStatusMessage method displays a message in the
   * Status Messages text area, followed by a special string.
   *
   * @param message  The message to be displayed in the Status
   *                 Messages text area.
   * @param special  The special string to be output to the Status
   *                 Messages text area.  For example, if no
   *                 linefeed is desired after message, special
   *                 would be null.
   */
  public synchronized void displayStatusMessage (String Message, String Special)
  {
    textArea_StatusMessages.appendText (Message + Special);
  }

  /**
   * The displayError method displays an error message in the
   * Status Messages text area.  The Bell sound is played
   * to indicate an error has occurred.
   *
   * @param ErrorMsg  The error message to be displayed in the
   *                  Status Messages text area.
   */
  public synchronized void displayError (String ErrorMsg)
  {
    // Play the default Bell sound.
    Bell.play();

    // Print the error message in the Status Messages text area.
    textArea_StatusMessages.appendText ("Error: " + ErrorMsg + "\n");
  }

  /**
   * The GetSNEidFromSNEAttrList method gets the SNEid from
   * the first entry in the list_SNEAttributes list.  If the
   * text on the first line that is printed to the list by
   * the buttonGetAttributes_Clicked method is changed, this
   * method will have to be updated.  If the list is empty,
   * null is returned, otherwise the SNEid string is returned.
   */
  public String GetSNEidFromSNEAttrList()
  {
    if (list_SNEAttributes.countItems() > 0)
    {
      String item = list_SNEAttributes.getItem (0);           // Get first list item line.
      StringTokenizer st = new StringTokenizer (item);        // The 4th token in the field
      for (int i = 1; i <= 4; i++) { item = st.nextToken(); } // is the SNE ID.
      String SNEid = item.substring (0, item.indexOf (':'));  // Remove the ":" at the end.
      return SNEid;
    }
    else
    {
      return "";
    }
  }

  /**
   * The GetRemoteSNEAttributes method gets remote attributes
   * from a remote object.  In the context of the simulated
   * environment, this remote object will usually be a SNE,
   * although it could be another object that supports the
   * SNE communication protocol.
   *
   * @param realIP    The IP address of the remote object.
   * @param realPort  The port that the remote object is listening on.
   */
  public void GetRemoteSNEAttributes (String realIP, int realPort)
  {
    RemoteAttributes.clear();  // Clear the Hashtable to make it ready for new data.

    // This is the code that will actually retrieve the remote values.
    String sendData = "RECEIVE REMOTE ALL";
    SCPCommunicator theComm = new SCPCommunicator (RemoteAttributes, sendData,
                                                   realIP, realPort);
    try { theComm.join(); } catch (InterruptedException ie) {}
    if (! theComm.isError())
    {
      displayStatusMessage ("Remote attributes retrieved.");
    }
    else
    {
      displayStatusMessage ("");
      displayError ("Remote Attributes could not be retrieved.");
    }
  }

  /**
   * The SetRemoteSNEAttributes method sets remote attributes
   * on a remote object.  In the context of the simulated
   * environment, this remote object will usually be a SNE,
   * although it could be another object that supports the
   * SNE communication protocol.
   *
   * @param realIP    The IP address of the remote object.
   * @param realPort  The port that the remote object is listening on.
   * @param key       The key to set in the remote object's table.
   * @param value     The value corresponding to the key.
   */
  public boolean SetRemoteSNEAttributes (String realIP, int realPort,
                                         String key, String value)
  {
    // This is the code that will actually set the remote value.
    String sendData = "SHIP " + key + " " + value;
    SCPCommunicator theComm = new SCPCommunicator (RemoteAttributes, sendData,
                                                   realIP, realPort);
    try { theComm.join(); } catch (InterruptedException ie) {}
    if (! theComm.isError())
    {
      displayStatusMessage ("Remote attribute set.");
      return true;
    }
    else
    {
      displayError ("Remote attribute could not be set.");
      return false;
    }
  }

  /**
   * The DeleteRemoteSNEAttributes method deletes remote attributes
   * from a remote object.  In the context of the simulated
   * environment, this remote object will usually be a SNE,
   * although it could be another object that supports the
   * SNE communication protocol.
   *
   * @param realIP    The IP address of the remote object.
   * @param realPort  The port that the remote object is listening on.
   * @param key       The key to delete in the remote object's table.
   */
  public boolean DeleteRemoteSNEAttributes (String realIP, int realPort, String key)
  {
    // This is the code that will actually delete the remote value.
    String sendData = "SHIP REMOVE " + key;
    SCPCommunicator theComm = new SCPCommunicator (RemoteAttributes, sendData,
                                                   realIP, realPort);
    try { theComm.join(); } catch (InterruptedException ie) {}
    if (! theComm.isError())
    {
      displayStatusMessage ("Remote attribute deleted.");
      return true;
    }
    else
    {
      displayError ("Remote attribute could not be deleted.");
      return false;
    }
  }

//  ------------------------------------------------------------------------------
  /**
   * The following methods do some processing in response
   * to a KeyPress event.  If the TAB key is pressed,
   * request that the focus be moved to the next textField.
   *
   * @param event  The Event object containing information
   *               about the event that has occurred.
   */

	void textFieldSNEID_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
	    textField_RealIP.requestFocus();
		}
	}

	void textFieldRealIP_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  textField_RealPort.requestFocus();
		}
	}

	void textFieldRealPort_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  textField_SimIP.requestFocus();
		}
	}

	void textFieldSimulatedIP_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
	    textField_SimPort.requestFocus();
	  }
	}

	void textFieldSimulatedPort_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  list_SNE.requestFocus();
		}
	}

	void listSNE_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  list_SNEAttributes.requestFocus();
		}
	}

	void listSNEAttributes_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  textField_Key.requestFocus();
		}
	}

	void textFieldKey_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  textField_Value.requestFocus();
		}
	}

	void textFieldValue_KeyPress(Event event) {
		if (event.key == TAB_KEY)
		{
		  textField_SNEID.requestFocus();
		}
	}

/*  ------------------------------------------------------------------------------
    The following methods handle button pressed events.
 */

  /**
   * The buttonCreate_Clicked method is called when the
   * Create button is clicked, checks that the SNEID,
   * RealIP and RealPort fields contain data, then
   * proceeds to enter the data into the SNE Database.
   * (SNED)
   * <p>
   * Note that the contents of the fields are not checked,
   * so no error is indicated for invalid IP addresses or
   * port numbers below 1024.  Also, no error is indicated
   * when real IP + port duplicates exist in the SNED.  An
   * error is indicated when the SNE already exists in the
   * SNED.
   *
   * @param event  The Event object containing information
   *               about the event that has occurred.
   */
	public void buttonCreate_Clicked (Event event) {
	  // Get values from the text fields.
	  String SNEID = textField_SNEID.getText();
	  String SimIP = textField_SimIP.getText();
	  int SimPort = 0;
	  try { SimPort = Integer.parseInt (textField_SimPort.getText()); }
	  catch (NumberFormatException nfe)
	  {
  	  displayError ("Real Port field is empty or does not contain an integer value.");
	  }
	  String RealIP = textField_RealIP.getText();
	  int RealPort = 0;
	  try { RealPort = Integer.parseInt (textField_RealPort.getText()); }
	  catch (NumberFormatException nfe)
	  {
  	  displayError ("Real Port field is empty or does not contain an integer value.");
	  }
	  if (SNEID.length() == 0)
	  {
  	  displayError ("SNE ID field must be filled in.");
	  }
	  else if (RealIP.length() == 0)
	  {
  	  displayError ("Real IP field must be filled in.");
	  }

	  else
	  // The SNEID, RealIP and RealPort fields contain data.
	  // Now proceed to create the SNE in the SNED.
	  {
	    if (theSned.CreateSNEentry (SNEID, SimIP, SimPort, RealIP, RealPort))
	    {
	      list_SNE.addItem (SNEID + ", " + RealIP + ", Listening on port " + RealPort);
	      displayStatusMessage ("Created SNE " + SNEID + " at IP " + RealIP +
	                            " on port " + RealPort + " (Simulated IP = " +
	                            SimIP + ", Simulated Port = " + SimPort + ").");
      }
      else
      {
        displayError ("SNE with SNE ID = " + SNEID + " already exists in the database.");
      }
	  }
	}

  /**
   * The buttonLink_Clicked method is called when the
   * Link button has been clicked.  It gets the
   * SNEid's from the list_SNE list, and calls the
   * CreateLink method in the sned class to create the
   * link in the SNED.  Note that this is all that has
   * to be done to set up a link; remote objects like
   * the SNE may periodically "call-in" to query the
   * link status from the SNED.  By default, the initial
   * link status is "UP".
   * <p>
   * If two items in the SNE List are not selected, an
   * error message is printed.
   *
   * @param event  The Event object containing information
   *               about the event that has occurred.
   */
	public void buttonLink_Clicked (Event event) {
	  int[] indexes = list_SNE.getSelectedIndexes();
	  if (indexes.length != 2)
	  {
  	  displayError ("Exactly two SNE's must be selected for Link - no more, no less.");
	  }
	  else
	  {
	    String[] items = list_SNE.getSelectedItems();
	    // Note that the items in the list must be deleted in
	    // reverse order, since deleting item 1 first will
	    // throw off all other items by -1.
      String fromSNEid = items[0].substring (0, items[0].indexOf (','));
      String toSNEid = items[1].substring (0, items[1].indexOf (','));
      theSned.CreateLink (fromSNEid, toSNEid, "UP");
	    displayStatusMessage (fromSNEid + " linked to " + toSNEid + " with link status of \"UP\".");
	    list_SNEAttributes.clear();  // Clear the Attributes list if links are added.
	  }
	}

  /**
   * The buttonDestroy_Clicked method is called when
   * the Destroy button has been clicked.  It deletes
   * all of the corresponding SNE's from the SNED and
   * removes their list entries from the list_SNE list.
   * <p>
   * If no items are selected, an error message is printed.
   *
   * @param event      The Event object containing information
   *                   about the event that has occurred.
   */
	public void buttonDestroy_Clicked (Event event) {
	  int[] indexes = list_SNE.getSelectedIndexes();
	  if (indexes.length < 1)
	  {
  	  displayError ("Select at least one SNE to be deleted.");
	  }
	  else
	  {
	    String[] items = list_SNE.getSelectedItems();
	    // Note that the items in the list must be deleted in
	    // reverse order, since deleting item 1 first will
	    // throw off all other items by -1.
	    for (int i = indexes.length - 1; i >= 0; i--)
	    {
	      String SNEid = items[i].substring (0, items[i].indexOf (','));
	      theSned.DeleteSNEentry (SNEid);
        list_SNE.delItem (indexes[i]);
	      displayStatusMessage ("Deleted " + SNEid + ".");
	    }
	  }
	}

  /**
   * The buttonGetAttributes_Clicked method is called when
   * the GetAttributes button has been clicked, or another
   * method calls it explicitly.  It gets both the local
   * attributes from the SNED and the remote attributes
   * from the SNE in another JVM, and displays these
   * attributes in the Attributes list.
   * <p>
   * Note that this method performs two functions, depending
   * upon parameters that are supplied to it by callers.<OL>
   * <LI> Obtain all attributes, including both local and
   *      remote attributes, and display them in the
   *      Attributes list (list_SNEAttributes).
   * <LI> Obtain only local attributes, use the remote
   *      attributes from a previous call, and display
   *      them in the Attributes list (list_SNEAttributes).</OL>
   * <p>
   * Function (1) is selected when the user explicitly hits
   * the Get Attributes button or double-clicks on an entry
   * in the list_SNE list.  Function (2) is selected when a
   * method has changed attributes only in the local SNED,
   * and wishes to show the user these attributes immediately
   * after the change (i.e. when a link is deleted). The
   * reason for function (2) is that local updates may happen
   * often, and to have the entire attribute list sent from
   * a remote object each time would be costly in terms of
   * network bandwidth.
   * <p>
   * If no items are selected, an error message is printed.
   *
   * @param event  The Event object containing information
   *               about the event that has occurred.
   * @param SNEid  The SNEid which is not null if function (2)
   *               is required.
   */
	public void buttonGetAttributes_Clicked (Event event, String SNEid) {
	  int index = list_SNE.getSelectedIndex();  // Get the index of the selected item.
	  if (index == -1 && SNEid.length() == 0)
	  // If there are no items or more than one item selected in the
	  // list, and function (1) is selected (meaning that an item
	  // MUST be selected since the user is explicitly requesting to
	  // get attributes), display an error, clear the Attributes
	  // list and exit the method.
	  {
  	  displayError ("Select one (and ONLY one) SNE to display.");
  	  list_SNEAttributes.clear();
	  }
	  else
	  {
	    // This flag is required to get around a sequence problem.
	    // If the SNEid was not provided by the caller, the Hashtable
	    // of remote attributes should be updated (see below for why).
	    // In order to update remote attributes, the realIP and realPort
	    // variables must be set.  However these variables must be set
	    // after the SNEid is set, which is where the remote attributes
	    // need to be updated.  Confused?  A simpler explanation would
	    // be that we need to defer the remote attribute update until
	    // all of its dependencies are taken care of, so we use a flag.
      boolean GetRemoteAttributesFlag = false;

	    if (SNEid.length() == 0)
  	  // If the SNEid was not provided by the caller, assume that
  	  // an item in the list_SNE was selected and the user explicitly
  	  // clicked the Get Attributes button or double-clicked on the
  	  // SNE item.  If this is the case, the remote attributes should
  	  // also be retrieved.  If the SNEid was provided by the caller,
  	  // the list should be updated as-is, without updating remote
  	  // data.
	    {
  	    String item = list_SNE.getSelectedItem();
    	  SNEid = item.substring (0, item.indexOf (','));
        GetRemoteAttributesFlag = true;  // Defer the remote attribute update.
  	  }

  	  // Get local SNE attributes from the SNED.
  	  String realIP = (String)theSned.GetSNEData (SNEid, "realIP");
  	  int realPort = ((Integer)theSned.GetSNEData (SNEid, "realPort")).intValue();
  	  String simIP = (String)theSned.GetSNEData (SNEid, "simIP");
  	  int simPort = ((Integer)theSned.GetSNEData (SNEid, "simPort")).intValue();

      if (GetRemoteAttributesFlag)
   	  // If we said before that we wanted to update the remote
   	  // attributes, then get the remote SNE attributes.
      {
        displayStatusMessage ("Retrieving Remote Attributes for " + SNEid + " ... ", "");
    	  GetRemoteSNEAttributes (realIP, realPort);
    	}

      // Get a copy of the links Hashtable.
  	  Hashtable links = theSned.GetLinks (SNEid);

      // Now add the links, local and remote attributes to the list.
      list_SNEAttributes.clear();  // Clear any old data from the list.

      list_SNEAttributes.addItem ("Local attributes for " + SNEid + ":");  // Area title.

      // Note that attributes are preceded by two spaces in the list.
      // This is done to distinguish them from area titles and
      // spacers.  These two spaces are VERY IMPORTANT to other
      // methods.  If this is changed, the other methods will have
      // to be updated accordingly.
      list_SNEAttributes.addItem ("  simIP    : " + simIP);
      list_SNEAttributes.addItem ("  simPort  : " + simPort);
      list_SNEAttributes.addItem ("  realIP   : " + realIP);
      list_SNEAttributes.addItem ("  realPort : " + realPort);
      list_SNEAttributes.addItem ("");  // Spacer.

      list_SNEAttributes.addItem ("Links for " + SNEid + ":");  // Area title.

      // Now we need to dump the contents of the links to a list.
      // This is done by using the keys method, which converts the
      // Hashtable keys into an enumerated object.  Each key and
      // associated value is output to the list, in the reverse
      // order that they were added to the Hashtable.
      Enumeration linkKeys = links.keys();
      while (linkKeys.hasMoreElements())
      {
        String key = (String)linkKeys.nextElement();
        String value = (String)links.get (key);
        list_SNEAttributes.addItem ("  " + key + " : " + value);
      }
      list_SNEAttributes.addItem ("");  // Spacer.

      list_SNEAttributes.addItem ("Remote attributes for " + SNEid +
                                  " (click here to add Key & Value):");  // Area title.

      // Now we need to dump the contents of the RemoteAttributes
      // Hashtable to a list, using the enumerated method
      // described previously.
      Enumeration remoteKeys = RemoteAttributes.keys();
      while (remoteKeys.hasMoreElements())
      {
        String key = remoteKeys.nextElement().toString();
        String value = RemoteAttributes.get (key).toString();
        list_SNEAttributes.addItem ("  " + key + " : " + value);
      }
      // Right after updating the Attributes list, nothing will
      // be selected so clear the key and value fields, and
      // disable everything.
      textField_Key.setText ("");
      textField_Value.setText ("");
      button_SetRemote.disable();
      textField_Key.setEditable (false);
      textField_Value.setEditable (false);
      button_Delete.disable();
      button_SetLocal.disable();
	  }
	}

  /**
   * The listSNE_Clicked method responds to a double-
   * click ACTION_EVENT by de-selecting all of the
   * items in the list except the item that was
   * double-clicked, then displaying the attributes
   * for that item.
   *
   * @param event      The Event object containing information
   *                   about the event that has occurred.
   */
	public void listSNE_Clicked (Event event) {
	  // Get the list item from the arg field of the event object.
    String item = (String)event.arg;

    // Deselect all list items by setting multiple selections
    // false then true.
    list_SNE.setMultipleSelections (false);
    list_SNE.setMultipleSelections (true);

    // Search through the list to find the item that was double-
    // clicked on.
    int index = 0;
    for (int i = 0; i < list_SNE.countItems(); i++)
    {
      if (item.equals (list_SNE.getItem (i)))
      {
        index = i;
        break;
      }
    }

    // Re-select the item that was double-clicked on, then call
    // the buttonGetAttributes_Clicked method to display the
    // atributes for the selected item.
    list_SNE.select (index);
		buttonGetAttributes_Clicked(event, "");
  }

  /**
   * The buttonSetRemote_Clicked method is called when
   * the Set Remote button has been clicked.  It gets
   * the text from the Key and Value text fields, and
   * updates a remote database with these new values by
   * calling the SetRemoteSNEAttributes method.  Note
   * that the Attributes list is cleared immediately
   * after this method executes, which will force the
   * user to manually re-display the remote attributes.
   * This is an extra confirmation step that will ensure
   * that the attribute was successfully transported to
   * its remote destination.
   * <p>
   * If the Key text field is empty, an error message is
   * displayed.
   *
   * @param event  The Event object containing information
   *               about the event that has occurred.
   */
	public void buttonSetRemote_Clicked(Event event) {
	  // Get the SNEid from the SNE Attributes list.
    String SNEid = GetSNEidFromSNEAttrList();

    int index = list_SNEAttributes.getSelectedIndex();

    String key = textField_Key.getText();
    String value = textField_Value.getText();
    if (key.length() != 0)
    {
      String realIP = (String)theSned.GetSNEData (SNEid, "realIP");
      int realPort = ((Integer)theSned.GetSNEData (SNEid, "realPort")).intValue();
      if (SetRemoteSNEAttributes (realIP, realPort, key, value))
      {
        displayStatusMessage ("Remote attribute set for " + SNEid + ": key " + key +
                              " set to " + value);
      }
      // Clear the Attributes list, forcing the user to manually
      // update the remote attributes.
      list_SNEAttributes.clear();
    }
    else
    {
      displayError ("Remote attribute not set because Key text field empty.");
    }
	}

  /**
   * The buttonSetLocal_Clicked method is called when
   * the Set Local button has been clicked.  It gets
   * the text from the Key and Value text fields, and
   * updates the SNED with these new values by calling
   * the appropriate method.  Note that the Attributes
   * list is updated immediately after this command
   * executes.
   * <p>
   * There are two areas in the Attributes list that are
   * affected by a Set Local button click: the Local
   * and Links areas.  Some calculations are done to
   * determine which area the selection falls under,
   * then appropriate action is taken.
   *
   * @param event      The Event object containing information
   *                   about the event that has occurred.
   */
	public void buttonSetLocal_Clicked(Event event) {
	  // Get the SNEid from the SNE Attributes list.
    String SNEid = GetSNEidFromSNEAttrList();

    int index = list_SNEAttributes.getSelectedIndex();

    // Search through the list to identify the two local
    // areas of the list: Local and Links.  It will be
    // important to know where the data is located when we
    // need to set the value.
    int local = 0;  // This is true since it's at the top of the list.
    int links = 0;
    for (int i = 2; i < list_SNEAttributes.countItems(); i++)
    // Note the loop starts at 2 because we are not interested
    // in the first two items in the list, which are the Local
    // title area (index known) and a possible spacer entry.
    {
      if (list_SNEAttributes.getItem (i).startsWith("Links"))
      {
        links = i;
        break;
      }
    }

    if (index < links)
    // If the item is in the Local area, set the attribute in
    // the SNED with the selected key and value.
    {
      String key = textField_Key.getText();
      String value = textField_Value.getText();
      theSned.SetSNEData (SNEid, key, value);
      displayStatusMessage ("Local attribute set for " + SNEid + ": key " + key +
                            " set to " + value);
    }

    if (index >= links)
    // If the item is in the Links area, set the link status
    // as required.
    {
      String toSNEid = textField_Key.getText();
      String value = textField_Value.getText();

      if (theSned.SetLinkStatus (SNEid, toSNEid, value))
      // If the SetLinkStatus method returns false, an undefined link
      // status was specified, so display an error message.  Otherwise
      // everything is okay and a status message is displayed.  For
      // a list of valid link statuses, see sne.java.
      {
        displayStatusMessage ("Link status between " + SNEid + " and " +
                              toSNEid + " set to " + value);
      }
      else
      {
        displayError ("Link could not be modified.  Undefined link status: " + value);
      }
    }
    buttonGetAttributes_Clicked (event, SNEid);
	}

  /**
   * The buttonDelete_Clicked method is called when
   * the Delete button has been clicked.  It gets
   * the SNEid from the Attributes list, gets the key
   * from the Key text field, and calls the appropriate
   * method to delete the item from the database.
   * <p>
   * There are two areas in the Attributes list that are
   * affected by a Delete button click: the Links
   * and Remote areas.  Some calculations are done to
   * determine which area the selection falls under,
   * then appropriate action is taken.
   *
   * @param event      The Event object containing information
   *                   about the event that has occurred.
   */
	public void buttonDelete_Clicked(Event event) {
	  // Get the SNEid from the SNE Attributes list.
    String SNEid = GetSNEidFromSNEAttrList();

    int index = list_SNEAttributes.getSelectedIndex();

    // Search through the list to identify the relevant
    // areas of the list: Links and Remote.  It will be
    // important to know where the data is located when we
    // need to delete the attribute.
    int links = 0;
    int remote = 0;
    for (int i = 2; i < list_SNEAttributes.countItems(); i++)
    // Note the loop starts at 2 because we are not interested
    // in the first two items in the list, which are the Local
    // title area (index known) and a possible spacer entry.
    {
      if (list_SNEAttributes.getItem (i).startsWith("Links"))
      {
        links = i;
      }
      if (list_SNEAttributes.getItem (i).startsWith("Remote"))
      {
        remote = i;
      }
    }

    // If the item is in the Links area, delete the link.
    if (index >= links && index < remote)
    {
      String toSNEid = textField_Key.getText();
      theSned.DeleteLink (SNEid, toSNEid);
      displayStatusMessage ("Deleted link between " + SNEid + " and " + toSNEid);
      buttonGetAttributes_Clicked (event, SNEid);
    }

    // If the item is in the Remote area, delete the
    // specified attribute.
    if (index >= remote)
    {
      String key = textField_Key.getText();
      String realIP = (String)theSned.GetSNEData (SNEid, "realIP");
      int realPort = ((Integer)theSned.GetSNEData (SNEid, "realPort")).intValue();
      if (DeleteRemoteSNEAttributes (realIP, realPort, key))
      {
        displayStatusMessage ("Deleted remote attribute from " + SNEid + " with key " + key);
      }
      list_SNEAttributes.clear();
    }
	}

/*  ------------------------------------------------------------------------------
    The following methods handle list select events.
 */

  /**
   * The listSNEAttributes_ListSelect method is called when
   * the list_SNEAttributes list has been selected.  This
   * means that the user has clicked on an item in the
   * Attributes list.  There are many possible actions for
   * this event.
   * <p>
   * The sole purpose of this method is to enable and/or
   * disable one or more of the Set Remote, Set Local,
   * and Delete buttons, and the Key and Value text fields.
   * Also, the Key and Value text fields are set to the key
   * and value as selected from the list, as appropriate.
   * <p>
   * An important point to note is that if no item in the
   * Attribute list is selected, all of the aforementioned
   * objects are disabled.  In order to enable selected
   * objects, the user must click on an attribute in one
   * of the three areas (Local, Links and Remote).  In the
   * case where there are no entries in the Remote area,
   * the user should click on the area title in order to
   * enable Key & Value text fields (as well as the Set
   * Remote and Delete buttons).  This will allow the user
   * to set new attributes on remote objects.
   *
   * @param event      The Event object containing information
   *                   about the event that has occurred.
   */
	public void listSNEAttributes_ListSelect(Event event) {
    int index = ((Integer)event.arg).intValue();       // Get index and string value
    String item = list_SNEAttributes.getItem (index);  // of selected item in list.

    // Search through the list to identify the three areas
    // of the list: Local, Links and Remote.  It will be
    // important to know where the data is located when we
    // enable/disable objects.
    int local = 0;  // This is true since it's at the top of the list.
    int links = 0;
    int remote = 0;
    for (int i = 2; i < list_SNEAttributes.countItems(); i++)
    // Note the loop starts at 2 because we are not interested
    // in the first two items in the list, which are the Local
    // title area (index known) and a possible spacer entry.
    {
      if (list_SNEAttributes.getItem (i).startsWith("Links"))
      {
        links = i;
      }
      if (list_SNEAttributes.getItem (i).startsWith("Remote"))
      {
        remote = i;
      }
    }

    // Since we now know what areas the attributes fall under
    // (one of Local, Links or Remote) by their position index
    // in the list, we can enable/disable objects as required.

    // Now enable/disable buttons and text fields based on the
    // location of the item in the list.
    if (item.startsWith ("  "))
    {
      // If the item is in the Local area, disable the SetRemote
      // and Delete buttons, and the Key text field.  Enable the
      // SetLocal button and the Value text field.
      if (index < links)
      {
        button_SetRemote.disable();
        button_Delete.disable();
        textField_Key.setEditable (false);
        textField_Value.setEditable (true);
        button_SetLocal.enable();
      }

      // If the item is in the Links area, disable the SetRemote
      // button, and the Key text field.  Enable the Delete and
      // SetLocal buttons, and the Value text field.
      if (index < remote && index >= links)
      {
        button_SetRemote.disable();
        textField_Key.setEditable (false);
        textField_Value.setEditable (true);
        button_Delete.enable();
        button_SetLocal.enable();
      }

      // If the item is in the Remote area, enable the SetRemote
      // and Delete buttons, and the Key text field.  Disable
      // the SetLocal button and enable the Value text field.
      if (index >= remote)
      {
        button_SetRemote.enable();
        textField_Key.setEditable (true);
        textField_Value.setEditable (true);
        button_Delete.enable();
        button_SetLocal.disable();
      }

      // Get the key and value using the string tokenizer, and
      // update the Key and Value text fields accordingly.
      StringTokenizer st = new StringTokenizer(item, " : ");
      String key = st.nextToken();
      textField_Key.setText (key);
      String value = st.nextToken();
      while (st.hasMoreTokens())
      // If the attribute has spaces in it, add them to the
      // value string.
      {
        value = value + " " + st.nextToken();
      }
      textField_Value.setText (value);
    }
    else if (item.startsWith ("Remote"))
    // If the Remote area title has been selected, enable the
    // SetRemote and Delete buttons, and the Key text field.
    // Disable the SetLocal button and enable the Value text
    // field.  Clear Key and Value text fields.
    {
      textField_Key.setText ("");
      textField_Value.setText ("");
      button_SetRemote.enable();
      textField_Key.setEditable (true);
      textField_Value.setEditable (true);
      button_Delete.enable();
      button_SetLocal.disable();
    }
    else
    // If an area title of blank line has been selected,
    // and it is not the Remote area title, clear the fields
    // and disable everything.
    {
      textField_Key.setText ("");
      textField_Value.setText ("");
      button_SetRemote.disable();
      textField_Key.setEditable (false);
      textField_Value.setEditable (false);
      button_Delete.disable();
      button_SetLocal.disable();
    }
	}

/*  ------------------------------------------------------------------------------
 */

  /**
   * The init method initializes and draws all applet controls in the
   * applet window.
   * <p>
   * This method is automatically created and updated by Visual Cafe.
   */
	public void init() {
		super.init();

		//{{INIT_CONTROLS
		setLayout(null);
		addNotify();
		resize(641,618);
		setFont(new Font("Dialog", Font.PLAIN, 10));
		setForeground(new Color(0));
		setBackground(new Color(12632256));
		label1 = new java.awt.Label("Simulated IP");
		label1.reshape(427,0,91,22);
		label1.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label1);
		label2 = new java.awt.Label("Real Port");
		label2.reshape(301,0,70,22);
		label2.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label2);
		label3 = new java.awt.Label("Status Messages");
		label3.reshape(0,458,119,22);
		label3.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label3);
		label4 = new java.awt.Label("SNE Manipulation Controls");
		label4.reshape(0,60,203,23);
		label4.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label4);
		label5 = new java.awt.Label("Current Simulated Network Elements (SNEs)");
		label5.reshape(0,113,301,23);
		label5.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label5);
		label6 = new java.awt.Label("SNE Attributes");
		label6.reshape(315,113,105,23);
		label6.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label6);
		label7 = new java.awt.Label("SNE ID");
		label7.reshape(0,0,63,22);
		label7.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label7);
		label8 = new java.awt.Label("Real IP");
		label8.reshape(175,0,77,22);
		label8.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label8);
		label9 = new java.awt.Label("Simulated Port");
		label9.reshape(539,0,98,22);
		label9.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label9);
		label11 = new java.awt.Label("Key");
		label11.reshape(315,390,49,22);
		label11.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label11);
		label12 = new java.awt.Label("Value");
		label12.reshape(315,420,49,22);
		label12.setFont(new Font("Dialog", Font.BOLD, 12));
		add(label12);
		button_Create = new java.awt.Button("Create");
		button_Create.reshape(21,83,84,22);
		add(button_Create);
		button_Link = new java.awt.Button("Link");
		button_Link.reshape(147,83,89,22);
		add(button_Link);
		button_Destroy = new java.awt.Button("Destroy");
		button_Destroy.reshape(280,83,84,22);
		add(button_Destroy);
		button_GetAttributes = new java.awt.Button("Get Attributes");
		button_GetAttributes.reshape(406,83,84,22);
		add(button_GetAttributes);
		button_ShipCode = new java.awt.Button("Ship Code");
		button_ShipCode.disable();
		button_ShipCode.reshape(532,83,84,22);
		add(button_ShipCode);
		button_SetRemote = new java.awt.Button("Set Remote");
		button_SetRemote.disable();
		button_SetRemote.reshape(364,450,77,22);
		add(button_SetRemote);
		button_SetLocal = new java.awt.Button("Set Local");
		button_SetLocal.disable();
		button_SetLocal.reshape(462,450,77,22);
		add(button_SetLocal);
		button_Delete = new java.awt.Button("Delete");
		button_Delete.disable();
		button_Delete.reshape(560,450,70,22);
		add(button_Delete);
		textField_SNEID = new java.awt.TextField();
		textField_SNEID.reshape(14,23,105,22);
		add(textField_SNEID);
		textField_SimIP = new java.awt.TextField();
		textField_SimIP.reshape(441,23,91,22);
		add(textField_SimIP);
		textField_SimPort = new java.awt.TextField();
		textField_SimPort.reshape(567,23,63,22);
		add(textField_SimPort);
		textField_RealPort = new java.awt.TextField();
		textField_RealPort.reshape(315,23,63,22);
		add(textField_RealPort);
		textField_RealIP = new java.awt.TextField();
		textField_RealIP.reshape(189,23,91,22);
		add(textField_RealIP);
		textField_Key = new java.awt.TextField();
		textField_Key.setEditable(false);
		textField_Key.reshape(364,390,266,23);
		add(textField_Key);
		textField_Value = new java.awt.TextField();
		textField_Value.setEditable(false);
		textField_Value.reshape(364,420,266,23);
		add(textField_Value);
		list_SNE = new java.awt.List(0,true);
		add(list_SNE);
		list_SNE.reshape(14,135,301,315);
		list_SNEAttributes = new java.awt.List(0,false);
		add(list_SNEAttributes);
		list_SNEAttributes.reshape(329,135,301,247);
		textArea_StatusMessages = new java.awt.TextArea();
		textArea_StatusMessages.setEditable(false);
		textArea_StatusMessages.reshape(14,480,616,135);
		add(textArea_StatusMessages);
		//}}

    // Initialize global data.
    theSned = new sned();  // The Simulated Network Element Database.
    RemoteAttributes = new Hashtable();  // The remote attribute Hashtable.

    // Get the audio clip to play as the Bell sound.  The file must be
    // in Sun audio format, it must be called Bell.au, and it must be
    // found in the same directory as the SCP applet.  If these
    // conditions are not met, no sound will be heard when the Bell clip
    // is played.
    try { Bell = getAudioClip (new URL (getCodeBase().toString() + "/Bell.au")); }
    catch (MalformedURLException mue) { }

    // Set up the Dispatcher to receive communication requests from other
    // network elements.
    Dispatcher theDispatcher = new Dispatcher (theSned, listenerPort,
                                               textArea_StatusMessages);
    displayStatusMessage ("Dispatcher thread started.  Listening on port " + listenerPort);
	}

  /**
   * The handleEvent method handles events.  When an event occurs, the
   * handleEvent method is called, and depending upon the object and
   * type of event, a specific event handler method is then called.
   * <p>
   * This method is automatically created and updated by Visual Cafe.
   *
   * @param event      The Event object containing information
   *                   about the event that has occurred.
   */
	public boolean handleEvent(Event event) {

	  // Handle all KEY_PRESS events.

		if (event.target == textField_SNEID && event.id == Event.KEY_PRESS) {
			textFieldSNEID_KeyPress(event);
		}
		if (event.target == textField_RealIP && event.id == Event.KEY_PRESS) {
			textFieldRealIP_KeyPress(event);
		}
		if (event.target == textField_RealPort && event.id == Event.KEY_PRESS) {
			textFieldRealPort_KeyPress(event);
		}
		if (event.target == textField_SimIP && event.id == Event.KEY_PRESS) {
			textFieldSimulatedIP_KeyPress(event);
		}
		if (event.target == textField_SimPort && event.id == Event.KEY_PRESS) {
			textFieldSimulatedPort_KeyPress(event);
		}
		if (event.target == list_SNE && event.id == Event.KEY_PRESS) {
			listSNE_KeyPress(event);
		}
		if (event.target == list_SNEAttributes && event.id == Event.KEY_PRESS) {
			listSNEAttributes_KeyPress(event);
		}
		if (event.target == textField_Key && event.id == Event.KEY_PRESS) {
			textFieldKey_KeyPress(event);
		}
		if (event.target == textField_Value && event.id == Event.KEY_PRESS) {
			textFieldValue_KeyPress(event);
		}

	  // Handle all ACTION_EVENT events.

		if (event.target == button_Create && event.id == Event.ACTION_EVENT) {
			buttonCreate_Clicked(event);
		}
		if (event.target == button_Link && event.id == Event.ACTION_EVENT) {
			buttonLink_Clicked(event);
		}
		if (event.target == button_Destroy && event.id == Event.ACTION_EVENT) {
			buttonDestroy_Clicked(event);
		}
		if (event.target == button_GetAttributes && event.id == Event.ACTION_EVENT) {
			buttonGetAttributes_Clicked(event, "");
		}
		if (event.target == button_SetRemote && event.id == Event.ACTION_EVENT) {
			buttonSetRemote_Clicked(event);
		}
		if (event.target == button_SetLocal && event.id == Event.ACTION_EVENT) {
			buttonSetLocal_Clicked(event);
		}
		if (event.target == button_Delete && event.id == Event.ACTION_EVENT) {
			buttonDelete_Clicked(event);
		}
		if (event.target == list_SNE && event.id == Event.ACTION_EVENT) {
			listSNE_Clicked(event);
		}

	  // Handle all LIST_SELECT events.

		if (event.target == list_SNEAttributes && event.id == Event.LIST_SELECT) {
			listSNEAttributes_ListSelect(event);
		}

		return super.handleEvent(event);
	}

	//{{DECLARE_CONTROLS
	java.awt.Label label1;
	java.awt.Label label2;
	java.awt.Label label3;
	java.awt.Label label4;
	java.awt.Label label5;
	java.awt.Label label6;
	java.awt.Label label7;
	java.awt.Label label8;
	java.awt.Label label9;
	java.awt.Label label11;
	java.awt.Label label12;
	java.awt.Button button_Create;
	java.awt.Button button_Link;
	java.awt.Button button_Destroy;
	java.awt.Button button_GetAttributes;
	java.awt.Button button_ShipCode;
	java.awt.Button button_SetRemote;
	java.awt.Button button_SetLocal;
	java.awt.Button button_Delete;
	java.awt.TextField textField_SNEID;
	java.awt.TextField textField_SimIP;
	java.awt.TextField textField_SimPort;
	java.awt.TextField textField_RealPort;
	java.awt.TextField textField_RealIP;
	java.awt.TextField textField_Key;
	java.awt.TextField textField_Value;
	java.awt.List list_SNE;
	java.awt.List list_SNEAttributes;
	java.awt.TextArea textArea_StatusMessages;
	//}}
}
