/**
 * @(#)sne.java
 *
 * Package: simu.sned
 *
 * This file contains the sne class, used by the
 * sned class as part of the Simulated Network
 * Element Database (SNED).  For more information,
 * see the sne class description.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970227   Andre Campeau   Creation.
 *
 */


package simu.sned;


import java.util.Hashtable;


/**
 * @version        1.0 27-Feb-97
 * @author         Andre Campeau
 * </DL><p>
 * The sne class contains information about Simulated Network
 * Elements (SNEs).  It is used as a component of the SNED
 * (Simulated Network Element Database - see sned.java).  Each
 * SNE defined in the SNED has a corresponding sne object
 * (a one-to-one relationship).
 * <p>
 * The methods of this class are called from other methods in
 * the sned class.  Thus the user (of a sned object) never
 * calls methods or accesses data directly within the sne class.
 * The methods are as follows:
 * <UL>
 * <LI> sne (nontrivial constructor)
 * <LI> get
 * <LI> put
 * <LI> getLink
 * <LI> getLinksTable
 * <LI> setLink
 * <LI> removeLink
 * </UL>
 * For more information, see the method descriptions included
 * with the methods.
 * <p>
 * The sne object is a Hashtable which contains several key-value
 * pairs that are relevant to SNE's.  The allowable data has been
 * hard-coded in the methods, so the user cannot add new keys
 * or remove any of the standard keys from the Hashtable.  This
 * approach was taken to facilitate the modification of the data
 * set of which the SNE is composed.  Care must be taken to
 * ensure that methods are updated to reflect any changes to the
 * sne class data set (referred to as "pre-defined keys and
 * values" in the method descriptions below).
 * <p>
 * An important part of the sne class data set is the "links"
 * Hashtable.  Each SNE may be linked to one or more SNEs in
 * the simulated network.  The "links" key in the sne object
 * has a corresponding value which is a pointer to a links
 * Hashtable.  The data in the links Hashtable are only
 * accessible from the sne class methods.
 * <p>
 */
public class sne
{

  // ---------------------------------------------------------------------
  // Define constants.

  int SNE_HASHTABLE_SIZE = 15;  // This is the initial size of the sne
                                // Hashtable.  This number should be
                                // roughly the same as the number of
                                // pre-defined keys in the Hashtable.

  int LINKS_HASHTABLE_SIZE = 15;  // This is the initial size of the links
                                  // Hashtable.  This number should be set
                                  // to the maximum number of links a SNE
                                  // will likely have.

  // ---------------------------------------------------------------------

  // The Simulated Network Element Hashtable object.
  private Hashtable sneTable;

  /**
   * The sne constructor creates the SNE object and populates
   * all fields.  An empty links Hashtable is created.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The empty SNED Hashtable exists.
   *
   * @param SimIP The Simulated IP Address.
   * @param SimPort The Simulated port.
   * @param RealIP The Real IP Address.
   * @param RealPort The Real port.
   */
  public sne (String SimIP, int SimPort, String RealIP, int RealPort)
  {
    // The pre-defined keys in the sneTable are:<UL>
    //   simIP
    //   simPort
    //   realIP
    //   realPort
    //   links
    // Only these keys and their corresponding values are
    // currently allowed in the sneTable.
    sneTable = new Hashtable (SNE_HASHTABLE_SIZE);
    sneTable.put ("simIP", SimIP);
    Integer SimPortInteger = new Integer (SimPort);
    sneTable.put ("simPort", SimPortInteger);
    sneTable.put ("realIP", RealIP);
    Integer RealPortInteger = new Integer (RealPort);
    sneTable.put ("realPort", RealPortInteger);
    Hashtable links = new Hashtable (LINKS_HASHTABLE_SIZE);
    sneTable.put ("links", links);
  }

  /**
   * The get method returns a value from the sne Hashtable.
   * <p>
   * Preconditions:  The sne constructor has been called.<p>
   * Postconditions: The value corresponding to the key in
   *                 the sne Hashtable is returned as an
   *                 Object.  If the key does not exist in
   *                 the Hashtable, null is returned.  Note
   *                 that the links Hashtable should not be
   *                 directly accessible from callers, so null
   *                 is returned if it is asked for.
   *
   * @param key The Hashtable lookup key.
   */
  public Object get (String key)
  {
    // The links Hashtable should not be directly accessible
    // from callers, so return null if it is asked for.  If
    // other similar objects are added to the sneTable, they
    // will have to be excluded as well.
    if (key.equals ("links"))
    {
      return null;
    }
    else
    {
      return sneTable.get (key);
    }
  }

  /**
   * The put method changes a value in the sne Hashtable.
   * <p>
   * Preconditions:  The sne constructor has been called.<p>
   * Postconditions: The value is modified for one of the
   *                 pre-defined keys in the sne HashTable.
   *                 If the key parameter is not one of the
   *                 pre-defined keys, false is returned,
   *                 otherwise true is returned.
   * <p>
   * The pre-defined keys in the sneTable are:<UL>
   * <LI> simIP
   * <LI> simPort
   * <LI> realIP
   * <LI> realPort</UL>
   *
   * @param key The Hashtable key.
   * @param value The Hashtable value; an Object.
   */
  public boolean put (String key, Object value)
  {
    // Only allow one of the pre-defined keys to be added
    // to the sneTable.  Note that the links key cannot
    // have its value overwritten.
    if (key.equals ("simIP")   ||
        key.equals ("simPort") ||
        key.equals ("realIP")  ||
        key.equals ("realPort"))
    {
      sneTable.put (key, value);
      return true;
    }
    else
    {
      return false;
    }
  }

  /**
   * The getLink method returns the link status of the
   * specified link.
   * <p>
   * Preconditions:  The sne constructor has been called.<p>
   * Postconditions: The link status of the specified link
   *                 is returned.  If the link does not
   *                 exist, null is returned.
   *
   * @param SNEid The SNE identifier.
   */
  public String getLink (String SNEid)
  {
    Hashtable links = (Hashtable) sneTable.get ("links");
    return (String) links.get (SNEid);
  }

  /**
   * The getLinksTable method returns a copy of the links
   * Hashtable.
   * <p>
   * Preconditions:  The links Hashtable exists.<p>
   * Postconditions: A deep copy of the links Hashtable is
   *                 returned.
   * <p>
   * This method allows a caller to get access to a copy
   * of the links Hashtable, without allowing the caller
   * to modify the real links Hashtable, which would
   * violate data hiding principles.
   */
  public synchronized Hashtable getLinksTable()
  {
    Hashtable links = (Hashtable) sneTable.get ("links");
    return links;
  }

  /**
   * The setLink method sets the link status of the
   * specified link.
   * <p>
   * Preconditions:  The sne constructor has been called.<p>
   * Postconditions: The link status of the specified link
   *                 is set.  If the link does not exist,
   *                 the link is created in the Hashtable
   *                 with SNEid as the key and linkStatus
   *                 as the value.<BR>
   *                 If the link status is not "UP" or "DOWN",
   *                 false is returned, otherwise true is
   *                 returned.
   *
   * @param SNEid The SNE identifier.
   */
  public boolean setLink (String SNEid, String linkStatus)
  {
    if (linkStatus.toUpperCase().equals ("UP") ||
        linkStatus.toUpperCase().equals ("DOWN"))
    {
      Hashtable links = (Hashtable) sneTable.get ("links");
      links.put (SNEid, linkStatus);
      return true;
    }
    else
    {
      return false;
    }
  }

  /**
   * The removeLink method removes the specified link from
   * the links Hashtable.
   * <p>
   * Preconditions:  The sne constructor has been called.<p>
   * Postconditions: If the link exists in the links
   *                 Hashtable, it is removed and true is
   *                 returned, otherwise false is returned.
   *
   * @param SNEid The SNE identifier.
   */
  public boolean removeLink (String SNEid)
  {
    Hashtable links = (Hashtable) sneTable.get ("links");
    if (links.containsKey (SNEid))
    {
      links.remove (SNEid);
      return true;
    }
    else
    {
      return false;
    }
  }
}
