/**
 * @(#)sned.java
 *
 * Package: simu.sned
 *
 * This file contains the sned class, used for the
 * Simulated Network Element Database (SNED).  For
 * more information, see the sned class description.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970227   Andre Campeau   Creation.
 *
 */


package simu.sned;


import java.util.Hashtable;
import java.util.Enumeration;
import simu.sned.sne;


/**
 * @version        1.0 27-Feb-97
 * @author         Andre Campeau
 * </DL><p>
 * The sned class is a database class for SNE's.  It is an
 * abbreviation for Simulated Network Element Database (SNED).
 * It contains a Hashtable which is the root of the database
 * (snedRoot) and many methods to access and mutate data in
 * the database.
 * <p>
 * All of the methods are synchronized, which will offer mutual
 * exclusion protection on the database.  This means that only
 * one thread can be accessing or mutating data in the SNED at
 * a time.  The methods are as follows:
 * <UL>
 * <LI> CreateSNEentry
 * <LI> DeleteSNEentry
 * <LI> GetSNEData
 * <LI> SetSNEData
 * <LI> CreateLink
 * <LI> GetLinkStatus
 * <LI> GetLinks
 * <LI> SetLinkStatus
 * <LI> DeleteLink
 * </UL>
 * For more information, see the method descriptions included
 * with the methods.
 * <p>
 * Simulated Network Elements (SNE's) can be defined and are the
 * basic records for the database.  Each SNE entry has a unique
 * SNEid and associated pointer to a SNE object.  The SNE object
 * is another Hashtable (see sne.java) and it contains several
 * pre-defined data elements which are associated with SNE's.
 * For example, the SNE object contains addressing and link
 * information about the SNE.
 * <p>
 */
public class sned
{

  // ---------------------------------------------------------------------
  // Define constants.

  int SNED_HASHTABLE_SIZE = 25;  // This is the initial size of the sned
                                 // Hashtable.  For debugging purposes,
                                 // it should be set small.  In production,
                                 // it should be set to the maximum number
                                 // of SNEs that will be running on the
                                 // simulated network.

  // ---------------------------------------------------------------------

  // The Simulated Network Element Database root Hashtable object.
  private Hashtable snedRoot;

  /**
   * The sned constructor creates an empty SNED Hashtable object
   * on the heap.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The empty SNED Hashtable exists.
   */
  public sned()
  {
    snedRoot = new Hashtable(SNED_HASHTABLE_SIZE);
  }

  /**
   * The CreateSNEentry method creates a SNE entry in the SNED.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: The SNEid is created as an attribute in the
   *                 SNED Hashtable.  A SNE Data object is
   *                 created and a pointer to it is inserted as
   *                 the SNEids value in the SNED Hashtable.
   *                 An empty Link Hashtable object is created and
   *                 a pointer to it is inserted into the Links
   *                 field of the SNE Data object.  The SNE Data
   *                 object is populated with mandatory data:
   *                 simIp, simPort, realIP and realPort.  If the
   *                 SNE id is already in the SNED, false is
   *                 returned and nothing is done, otherwise the
   *                 data is added to the SNED and true is returned.
   *
   * @param SNEid The SNE identifier.
   * @param simIP The Simulated IP Address.
   * @param simPort The Simulated port.
   * @param realIP The Real IP Address.
   * @param realPort The Real port.
   */
  public synchronized boolean CreateSNEentry (String SNEid, String simIp,
                                              int simPort,  String realIP,
                                              int realPort)
  {
    if (snedRoot.containsKey (SNEid))
    {
      return false;
    }
    else
    {
      sne SNE = new sne (simIp, simPort, realIP, realPort);
      snedRoot.put (SNEid, SNE);
      return true;
    }
  }

  /**
   * The DeleteSNEentry method deletes a SNE entry from the SNED.
   * <p>
   * Preconditions:  The SNEid exists in the SNED Hashtable.<p>
   * Postconditions: If the SNEid exists, its Link Hashtable is
   *                 scanned and the DeleteLink method is invoked
   *                 for all links in the Hashtable. The SNEid is
   *                 removed from the SNED, which leaves the SNE
   *                 Data Object unreferenced.  Automatic garbage
   *                 collection will automatically deallocate
   *                 space used by the SNE Data object and Link
   *                 Hashtable.  If the SNEid does not exist in
   *                 the SNED Hashtable, nothing is done.
   *
   * @param SNEid The SNE identifier.
   */
  public synchronized void DeleteSNEentry (String fromSNEid)
  {
    if (snedRoot.containsKey (fromSNEid))
    {
      sne SNE = (sne)snedRoot.get (fromSNEid);  // Get the SNE corresponding to fromSNEid.
  	  Hashtable links =  SNE.getLinksTable();  // Get a copy of the links Hashtable.
      Enumeration linkKeys = links.keys();
      while (linkKeys.hasMoreElements())
      {
        String toSNEid = (String)linkKeys.nextElement();
        DeleteLink (fromSNEid, toSNEid);
      }
      snedRoot.remove (fromSNEid);
    }
  }

  /**
   * The GetSNEData method gets SNE data from the SNE Data Object.
   * <p>
   * Preconditions:  The SNEid exists in the SNED Hashtable.<p>
   * Postconditions: If the SNEid exists, a lookup is done in its
   *                 SNE Data Object using the attribute parameter
   *                 and the resulting value is returned as a
   *                 String.  If SNEid does not exist in the SNED
   *                 Hashtable, or the attribute does not exist in
   *                 the SNE Data Object, or the value for the
   *                 attribute is set to the null string, a null
   *                 string is returned.
   *
   * @param SNEid The SNE identifier.
   * @param attribute The attribute (key) to be obtained.
   */
  public synchronized Object GetSNEData (String SNEid, String attribute)
  {
    if (snedRoot.containsKey (SNEid))
    {
      sne SNE = (sne) snedRoot.get (SNEid);
      return SNE.get (attribute);
    }
    else
    {
      return "";
    }
  }

  /**
   * The SetSNEData method creates and/or sets SNE data in the SNE Data Object.
   * <p>
   * Preconditions:  The SNEid exists in the SNED Hashtable.<p>
   * Postconditions: If the attribute exists in the SNE Data Object,
   *                 its value is changed to the new value.  If the
   *                 attribute does not exist, nothing is done.
   *                 If the SNEid does not exist in the SNED Hashtable,
   *                 nothing is done.
   *
   * @param SNEid The SNE identifier; a string.
   * @param attribute The attribute (key) to be inserted; a string.
   * @param value The data to be inserted; an object.
   */
  public synchronized void SetSNEData (String SNEid, String attribute, Object value)
  {
    if (snedRoot.containsKey (SNEid))
    {
      sne SNE = (sne) snedRoot.get (SNEid);
      SNE.put (attribute, value);
    }
  }

  /**
   * The CreateLink method creates a simulated network link by adding the
   * appropriate data to the SNED.
   * <p>
   * Preconditions:  Both fromSNEid and toSNEid exist in the SNED
   *                 Hashtable.<p>
   * Postconditions: The link is created in the SNED.  A Link Hashtable
   *                 object is populated with the link data.  Note that
   *                 both the fromSNEid and toSNEid Link Hashtables are
   *                 updated.  If the link already existed, the
   *                 Link Status is replaced by the contents of the
   *                 linkStatus parameter.  If one or both of the
   *                 SNEids do not exist in the SNED Hashtable,
   *                 nothing is done.
   *
   * @param fromSNEid The first SNE identifier.
   * @param toSNEid The second SNE identifier.
   * @param linkStatus The link status.
   */
  public synchronized void CreateLink (String fromSNEid, String toSNEid,
                                       String linkStatus)
  {
    SetLinkStatus (fromSNEid, toSNEid, linkStatus);
  }

  /**
   * The GetLinkStatus method retrieves the Link Status from the SNED.
   * <p>
   * Preconditions:  Both fromSNEid and toSNEid exist in the SNED
   *                 Hashtable.<p>
   * Postconditions: The link status is looked up in the Link
   *                 Hashtable and returned as a string.  If the
   *                 SNEid's do not exist in the SNED Hashtable,
   *                 a null string is returned.  If the link does
   *                 not exist, a null string is returned.
   *
   * @param fromSNEid The first SNE identifier.
   * @param toSNEid The second SNE identifier.
   */
  public synchronized String GetLinkStatus (String fromSNEid, String toSNEid)
  {
    if (snedRoot.containsKey (fromSNEid) && snedRoot.containsKey (toSNEid))
    {
      sne SNE = (sne) snedRoot.get (fromSNEid);  // Get the SNE corresponding to fromSNEid.
      return SNE.getLink (toSNEid);        // Return the link status.
    }
    else
    {
      return "";
    }
  }

 /**
   * The GetLinks method retrieves a list of links for a SNE.
   * <p>
   * Preconditions:  None.<p>
   * Postconditions: A list of links and the corresponding link
   *                 status is returned in a Hashtable.  If
   *                 there are no links for the SNE, or the
   *                 SNEid does not exist in the SNED, null is
   *                 returned.
   *
   * @param fromSNEid The first SNE identifier.
   * @param toSNEid The second SNE identifier.
   */
  public synchronized Hashtable GetLinks (String SNEid)
  {
    if (snedRoot.containsKey (SNEid))
    {
      sne SNE = (sne) snedRoot.get (SNEid);  // Get the SNE corresponding to SNEid.
  	  return SNE.getLinksTable();  // Return a copy of the links Hashtable.
    }
    else
    {
      return null;
    }
  }

  /**
   * The SetLinkStatus method sets the Link Status in the SNED.
   * <p>
   * Preconditions:  Both fromSNEid and toSNEid exist in the SNED
   *                 Hashtable.<p>
   * Postconditions: If the link exists, the Link Status is replaced
   *                 by the contents of the linkStatus parameter.
   *                 If one or both of the SNEs do not exist in the
   *                 SNED Hashtable, or the link does not exist,
   *                 false is returned, otherwise true is returned.
   *
   * @param fromSNEid The first SNE identifier.
   * @param toSNEid The second SNE identifier.
   * @param linkStatus The link status.
   */
  public synchronized boolean SetLinkStatus (String fromSNEid, String toSNEid,
                                          String linkStatus)
  {
    if (snedRoot.containsKey (fromSNEid) && snedRoot.containsKey (toSNEid))
    {
      sne SNE;  // Define SNE as a sne Object.

      SNE = (sne) snedRoot.get (fromSNEid);  // Get the SNE corresponding to fromSNEid.
      SNE.setLink (toSNEid, linkStatus);    // Add/update the link to the SNE.

      SNE = (sne) snedRoot.get (toSNEid);    // Get the SNE corresponding to toSNEid.
      return SNE.setLink (fromSNEid, linkStatus); // Add/update the link to the SNE.
    }
    else
    {
      return false;
    }
  }

  /**
   * The DeleteLink method deletes the link from the SNED.
   * <p>
   * Preconditions:  Both fromSNEid and toSNEid exist in the SNED
   *                 Hashtable.<p>
   * Postconditions: The link is removed from the Link Hashtable
   *                 for both fromSNEid and toSNEid.  If one or
   *                 both of the SNEs do not exist in the SNED
   *                 Hashtable, or the link does not exist, nothing
   *                 is done.
   *
   * @param fromSNEid The first SNE identifier.
   * @param toSNEid The second SNE identifier.
   */
  public synchronized void DeleteLink (String fromSNEid, String toSNEid)
  {
    if (snedRoot.containsKey (fromSNEid) && snedRoot.containsKey (toSNEid))
    {
      sne SNE;  // Define SNE as a sne Object.

      SNE = (sne) snedRoot.get (fromSNEid);  // Get the SNE corresponding to fromSNEid.
      SNE.removeLink (toSNEid);        // Remove the link from the SNE.

      SNE = (sne) snedRoot.get (toSNEid);    // Get the SNE corresponding to toSNEid.
      SNE.removeLink (fromSNEid);      // Remove the link from the SNE.
    }
  }
}