/**
 * @(#)TestCommunicator.java
 *
 * Package: simu.net
 *
 * This file contains the TestCommunicator class, which is a test
 * harness for the Communicator class.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970320   Andre Campeau   Creation.
 *
 */


package simu.test;


import java.net.ServerSocket;
import java.io.DataInputStream;
import java.io.IOException;
import java.lang.InterruptedException;
import java.util.Enumeration;
import java.util.Hashtable;
import simu.net.Communicator;


/**
 * @version        1.0 20-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The TestCommunicator class is a test harness for the Communicator
 * class.  Testing of both sending and receiving capabilities of the
 * Communicator is done.
 * <p>
 * @see simu.net.Communicator
 * @see simu.test.TestSCPCommunicator
 */
public class TestCommunicator
{

  /**
   * The Send method connects with a dispatcher on the local host
   * that is listening on a well-known port, receives the new
   * anonymous port from the dispatcher, and finally opens another
   * connection on this anonymous port to send data (and possibly
   * receive a reply).
   * <p>
   * This method exercises the dispatcher connection and send
   * capabilities of the Communicator class.
   * <p>
   * Preconditions:  A dispatcher on the local host is listening on
   *                 a well-known port.<p>
   * Postconditions: Data is sent (and any reply is received) to the
   *                 remote object over the socket connection.
   *
   * @param port           The well-known port that the dispatcher
   *                       is listening on.
   * @param sendData       The data to send to the remote end.
   * @param receiveData[]  Any received data.  This is an array of
   *                       size 1.
   */
  public static void Send (int port, String sendData, String receiveData[])
  {
    int sendPortArray[] = new int[1];

    // Set the port array to the dispatcher's well known port.
    sendPortArray[0] = port;

    // Open a connection to the dispatcher and get the anonymous
    // port number.  Wait for the connection to close.
    Communicator theComm = new Communicator ("127.0.0.1", sendPortArray);
    try { theComm.join(); } catch (InterruptedException ie) {}

    System.out.println ("TestCommunicator reconnecting on port " + sendPortArray[0]);

    // Open the connection to the SCPCommunicator which is
    // listening on the anonymous port and send the data.
    // Wait for the connection to close.
    theComm = new Communicator (receiveData, sendData, "127.0.0.1", sendPortArray[0]);
    try { theComm.join(); } catch (InterruptedException ie) {}
  }

  /**
   * The main method executes a number of test scenarios.  These
   * scenarios are:<OL>
   * <LI> Create a Communicator listener daemon thread and verify
   *      that it can receive data and update a local Hashtable.
   *      After the last test is executed, the contents of the
   *      local Hashtable are printed to the standard output to
   *      verify its contents.
   * <LI> Send the data "SHIP LINKS SNE1 SNE2 UP" to the SCP and
   *      verify that the data is entered into the SNED.
   * <LI> Send the data "RECEIVE LOCAL SNE1 simIP" to the SCP and
   *      verify that a reply is received by printing the simIP
   *      to the standard output.
   * <LI> Send the data "RECEIVE LINKS SNE1 SNE2" to the SCP and
   *      verify that a reply is received by printing the link
   *      status to the standard output.</OL>
   * <p>
   * This method exercises the dispatcher connection and send
   * capabilities of the Communicator class (by calling the Send
   * method), as well as the listener daemon capability of the
   * Communicator class.
   * <p>
   * Preconditions:  The Simulator Control Program (SCP) is running
   *                 on the local host with its Dispatcher listening
   *                 on a well-known port.<p>
   * Postconditions: The operation of the Communicator class is
   *                 verified.  The operation of the Dispatcher and
   *                 SCPCommunicator classes is also verified
   *                 indirectly.
   *
   * @param args[]  The listener port for the Communicator listener
   *                daemon and the port that the SCP's Dispatcher
   *                is listening on.
   */
  public static void main (String args[])
  {
    Hashtable testData = new Hashtable();
    String receiveData[] = new String[1];
    DataInputStream dis = new DataInputStream (System.in);

    if (args.length < 2)
    {
      System.out.println("USAGE: java TestCommunicator listenerPort SCPport");
      System.exit(1);
    }
    int listenerPort = Integer.parseInt (args[0]);
    int SCPport = Integer.parseInt (args[1]);

    // Test 1:  Create a Communicator listener daemon thread and
    // verify that it can receive data and update a local Hashtable.
    Communicator theComm = new Communicator (testData, listenerPort);

    // Test 2:  Send the data "SHIP LINKS SNE1 SNE2 UP" to the
    // SCP and verify that the data is entered into the SNED.
    try { dis.readLine(); } catch (IOException ioe) {}
    Send (SCPport, "SHIP LINKS SNE1 SNE2 UP", receiveData);

    // Test 3:  Send the data "RECEIVE LOCAL SNE1 simIP" to the
    // SCP and verify that a reply is received by printing the
    // simIP to the standard output.
    try { dis.readLine(); } catch (IOException ioe) {}
    Send (SCPport, "RECEIVE LOCAL SNE1 simIP", receiveData);
    System.out.println ("Receive data = " + receiveData[0]);

    // Test 4:  Send the data "RECEIVE LINKS SNE1 SNE2" to the
    // SCP and verify that a reply is received by printing the
    // link status to the standard output.</OL>
    try { dis.readLine(); } catch (IOException ioe) {}
    Send (SCPport, "RECEIVE LINKS SNE1 SNE2", receiveData);
    System.out.println ("Receive data = " + receiveData[0]);

    // Test 1 (continued):  After the last test is executed,
    // the contents of the local Hashtable are printed to
    // the standard output to verify its contents.
    System.out.println ("Printing testData contents:");
    Enumeration testKeys = testData.keys();
    while (testKeys.hasMoreElements())
    {
      String key = (String)testKeys.nextElement();
      String value = (String)testData.get (key);
      System.out.println ("  " + key + " : " + value);
    }
    while (true) {}
  }
}