/**
 * @(#)TestSCPCommunicator.java
 *
 * Package: simu.net
 *
 * This file contains the TestSCPCommunicator class, which
 * is a test harness for the SCPCommunicator class.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970320   Andre Campeau   Creation.
 *
 */


package simu.test;


import java.net.ServerSocket;
import java.io.DataInputStream;
import java.io.IOException;
import java.lang.InterruptedException;
import java.util.Hashtable;
import java.util.Enumeration;
import simu.sned.sned;
import simu.net.SCPCommunicator;


/**
 * @version        1.0 20-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The TestSCPCommunicator class is a test harness for the
 * SCPCommunicator class.  Testing of both sending and
 * receiving capabilities of the SCPCommunicator is done.
 * <p>
 * @see simu.net.SCPCommunicator
 * @see simu.test.TestCommunicator
 */
public class TestSCPCommunicator
{

  /**
   * The main method executes a number of test scenarios.  These
   * scenarios are:<OL>
   * <LI> Verify the SCPCommunicator class listener thread
   *      constructors by creating two listener threads.  First,
   *      create a SCPCommunicator listener daemon thread on port
   *      1100.  Through interaction with the TestCommunicator
   *      class test harness (or similar program), verify that it
   *      can send data from the SNED (Simulated Network Element
   *      Database), and receive data and update the SNED.
   * <LI> Second, create a SCPCommunicator listener daemon thread
   *      on port 1101, and verify the same operations as in the
   *      previous test.  Note that there are no commands to
   *      print the SNED contents to the standard output.  To
   *      verify the SNED contents for both this and the previous
   *      test, a debugger should be used to inpect the data.</OL>
   * <p>
   * Susequent tests exercise the send constructor for the
   * SCPCommunicator class.  Data is sent, and any replies
   * are returned in the local Hashtable. After the last test
   * is executed, the contents of the local Hashtable are
   * printed to the standard output to verify its contents.
   * <p><OL START=3>
   * <LI> Send the data "SHIP key5 value5" to a Communicator class
   *      (or equivalent) to set a remote attribute in the remote
   *      database.
   * <LI> Send the data "SHIP REMOVE key5" to verify that an
   *      attribute can be removed from the remote database.
   * <LI> Send the data "SHIP key6 value6" and "SHIP key7 value7"
   *      to set more remote attributes.
   * <LI> Send the data "RECEIVE key6" to verify that the value
   *      of the attribute corresponding to key6 can be retrieved
   *      from the remote object.
   * <LI> Send the data "RECEIVE REMOTE ALL" to verify that all
   *      of the remote attributes can be retrieved.</OL>
   * <p>
   * Preconditions:  A communicator thread, which adheres to the
   *                 Communicator class protocol, is listening on
   *                 port 1102.  Note that this test harness was
   *                 designed to run in conjunction with the
   *                 TestCommunicator class test harness, which
   *                 exercises the Communicator class.  In general,
   *                 however, any communicator thread that adheres
   *                 to the Communicator class protocol would be
   *                 sufficient.<p>
   * Postconditions: The operation of the SCPCommunicator class is
   *                 verified.
   *
   * @param args[]  No command line arguments are required.<p>
   * @see simu.net.Communicator
   */
  public static void main (String args[])
  {
    // Define variables and populate the SNED.
    DataInputStream dis = new DataInputStream (System.in);
    sned theSned = new sned();
    theSned.CreateSNEentry ("SNE1", "1.2.3.4", 2000, "134.117.1.1", 1101);
    theSned.CreateSNEentry ("SNE2", "1.2.3.5", 2000, "134.117.1.1", 1102);
    theSned.CreateLink ("SNE1", "SNE2", "DOWN");
    Hashtable remoteData = new Hashtable();

    // Test 1:  Create a SCPCommunicator listener daemon thread
    // on port 1100.  Through interaction with the TestCommunicator
    // class test harness (or similar program), verify that it
    // can send data from the SNED (Simulated Network Element
    // Database), and receive data and update the SNED.
    SCPCommunicator theComm = new SCPCommunicator (theSned, 1100);

    // Test 2:  Create a SCPCommunicator listener daemon thread
    // on port 1101, and verify the same operations as in the
    // previous test.
    try
    {
      ServerSocket listener = new ServerSocket (1101);
      theComm = new SCPCommunicator (theSned, listener);
    }
    catch (IOException ioe) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    // Test 3:  Send the data "SHIP key5 value5" to a Communicator
    // class (or equivalent) to set a remote attribute in the remote
    // database.
    String sendData = "SHIP key5 value5";
    theComm = new SCPCommunicator (remoteData, sendData, "127.0.0.1", 1102);
    try { theComm.join(); } catch (InterruptedException ie) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    // Test 4:  Send the data "SHIP REMOVE key5" to verify that an
    // attribute can be removed from the remote database.
    sendData = "SHIP REMOVE key5";
    theComm = new SCPCommunicator (remoteData, sendData, "127.0.0.1", 1102);
    try { theComm.join(); } catch (InterruptedException ie) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    // Test 5:  Send the data "SHIP key6 value6" and "SHIP key7 value7"
    // to set more remote attributes.
    sendData = "SHIP key6 value6";
    theComm = new SCPCommunicator (remoteData, sendData, "127.0.0.1", 1102);
    try { theComm.join(); } catch (InterruptedException ie) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    sendData = "SHIP key7 value7";
    theComm = new SCPCommunicator (remoteData, sendData, "127.0.0.1", 1102);
    try { theComm.join(); } catch (InterruptedException ie) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    // Test 6:  Send the data "RECEIVE key6" to verify that the value
    // of the attribute corresponding to key6 can be retrieved
    // from the remote object.
    sendData = "RECEIVE key6";
    theComm = new SCPCommunicator (remoteData, sendData, "127.0.0.1", 1102);
    try { theComm.join(); } catch (InterruptedException ie) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    // Test 7:  Send the data "RECEIVE REMOTE ALL" to verify that all
    // of the remote attributes can be retrieved.
    sendData = "RECEIVE REMOTE ALL";
    theComm = new SCPCommunicator (remoteData, sendData, "127.0.0.1", 1102);
    try { theComm.join(); } catch (InterruptedException ie) {}
    try { dis.readLine(); } catch (IOException ioe) {}

    // Print the contents of the local Hashtable to
    // the standard output to verify its contents
    // (applies to tests 3 through 7).
    System.out.println ("Printing remoteData contents:");
    Enumeration remoteKeys = remoteData.keys();
    while (remoteKeys.hasMoreElements())
    {
      String key = (String)remoteKeys.nextElement();
      String value = (String)remoteData.get (key);
      System.out.println ("  " + key + " : " + value);
    }

    while (true) {}
  }
}