/**
 * @(#)Semaphore.java
 *
 * Package: simu.util
 *
 * This file contains the Semaphore class, which is
 * used for inter-process-communication - specifically
 * the synchronization of threads.
 *
 * Revision History
 * ================
 *
 * DATE     CHANGED BY      DESCRIPTION
 * ----     ----------      -----------
 * 970327   Andre Campeau   Creation.
 *
 */


package simu.util;


/**
 * @version        1.0 27-Mar-97
 * @author         Andre Campeau
 * </DL><p>
 * The Semaphore class implements a simple counting semaphore.
 * All methods are synchronized, meaning that the Semaphore
 * class is actually a monitor.  Calling threads wait for access
 * to shared resources, which in this case is the semCount
 * class variable.
 * <p>
 * The methods are as follows:<BR>
 *  - Wait<BR>
 *  - Signal<BR>
 * For more information, see the method descriptions included
 * with the methods.
 * <p>
 * The behaviour of the Semaphore class is that of a counting
 * semaphore, which means that calls to Wait() will cause the
 * caller to block if the semCount is zero, and to decrement
 * the semCount if it is greater than zero.  Calls to Signal()
 * will increment the semCount.
 * <p>
 */
public class Semaphore
{
  private int semCount = 0;  // The semaphore count.

/**
 * This Semaphore class constructor is called to create an
 * instance of the semaphore class with semCount set initially
 * to zero.
 */
  public Semaphore() {}

/**
 * This Semaphore class constructor is called to create an
 * instance of the semaphore class with semCount set initially
 * to the value specified by the SemCount parameter.
 * @param SemCount The initial semaphore count.
 */
  public Semaphore (int SemCount)
  {
    semCount = SemCount;
  }

/**
 * The Wait method causes the caller to block if the semCount
 * is zero, and to decrement the semCount if it is greater
 * than zero.  The method is synchronized to provide mutual
 * exclusion protection.
 */
  public synchronized void Wait()
  {
    if (semCount == 0)
    {
      try { wait(); } catch (InterruptedException ie) {}
    }
    else
    {
      semCount--;
    }
  }

/**
 * The Signal method causes the semCount to be incremented.
 * If any threads are waiting for a signal, they will be
 * notified and moved to the ready-to-run queue.
 */
  public synchronized void Signal()
  {
    semCount++;
    notify();
  }
}